##
# Copyright (c) 2013-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from __future__ import print_function

"""
LDAP directory service tests.
"""

from itertools import chain

import ldap


try:
    from mockldap import MockLdap
    from mockldap.filter import (
        Test as MockLDAPFilterTest,
        UnsupportedOp as MockLDAPUnsupportedOp,
    )
    from mockldap.ldapobject import LDAPObject

    # Allow ldap.async to work with mockldap:

    def search_ext(
        self, base, scope, filterstr='(objectClass=*)',
        attrlist=None, attrsonly=0, serverctrls=None, clientctrls=None,
        timeout=-1, sizelimit=0
    ):
        value = self._search_s(base, scope, filterstr, attrlist, attrsonly)
        if sizelimit:
            value = value[:sizelimit]
        return self._add_async_result(value)


    def result3(self, msgid, all=1, timeout=None):
        return ldap.RES_SEARCH_RESULT, self._pop_async_result(msgid), None, None

    LDAPObject.search_ext = search_ext
    LDAPObject.result3 = result3

except ImportError:
    MockLdap = None

from twext.python.types import MappingProxyType
from twext.who.idirectory import RecordType
from twext.who.ldap import (
    LDAPAttribute, RecordTypeSchema, LDAPObjectClass
)
from twext.who.util import ConstantsContainer

from twisted.python.constants import Names, NamedConstant, ValueConstant
from twisted.python.filepath import FilePath
from twisted.internet.defer import inlineCallbacks
from twisted.cred.credentials import UsernamePassword
from twisted.trial import unittest
from twisted.trial.unittest import SkipTest

from ...idirectory import FieldName as BaseFieldName
from .._service import (
    DEFAULT_FIELDNAME_ATTRIBUTE_MAP, DEFAULT_RECORDTYPE_SCHEMAS,
    LDAPBindAuthError,
    DirectoryService, DirectoryRecord,
)

from ...test import test_directory
from ...test.test_xml import (
    xmlService,
    BaseTest as XMLBaseTest, QueryMixIn,
    DirectoryServiceConvenienceTestMixIn,
    DirectoryServiceRealmTestMixIn,
    DirectoryServiceQueryTestMixIn as BaseDirectoryServiceQueryTestMixIn,
    DirectoryServiceMutableTestMixIn as BaseDirectoryServiceMutableTestMixIn
)


class TestFieldWithChoices(Names):

    none = NamedConstant()
    none.description = u"none"

    one = NamedConstant()
    one.description = u"one"

    two = NamedConstant()
    two.description = u"two"

    three = NamedConstant()
    three.description = u"three"


class TestFieldName(Names):
    multiChoice = NamedConstant()
    multiChoice.description = u"Multiple Choice Test Field"
    multiChoice.valueType = TestFieldWithChoices

    boolean1 = NamedConstant()
    boolean1.description = u"Boolean Test Field"
    boolean1.valueType = bool

    boolean2 = NamedConstant()
    boolean2.description = u"Boolean Test Field 2"
    boolean2.valueType = bool

    boolean3 = NamedConstant()
    boolean3.description = u"Boolean Test Field 3"
    boolean3.valueType = bool


TEST_FIELDNAME_MAP = dict(DEFAULT_FIELDNAME_ATTRIBUTE_MAP)
TEST_FIELDNAME_MAP[BaseFieldName.uid] = (u"__who_uid__", u"altuid")

TEST_FIELDNAME_MAP[BaseFieldName.fullNames] = (u"altname", u"cn")

TEST_FIELDNAME_MAP[TestFieldName.multiChoice] = (
    u"testField:One:one",
    u"testField:Two:two",
    u"testField:Three:three",
)

# Set up two Fields which will map to the same LDAP attribute ("foo")
# to make sure both Fields get set.  Their field values *will* be different,
# based on whether there is a colon in the attribute rule.  If the LDAP
# value matches the string following the colon, the Field value will be True.
# If there is no colon, then the string to match is literally the string "true".
TEST_FIELDNAME_MAP[TestFieldName.boolean1] = (
    u"foo:active",
)
TEST_FIELDNAME_MAP[TestFieldName.boolean2] = (
    u"foo",
)

# This Field will map to LDAP attribute "bar" and will be used for matching
# the string "true"
TEST_FIELDNAME_MAP[TestFieldName.boolean3] = (
    u"bar",
)




class TestService(DirectoryService, QueryMixIn):
    pass



class BaseTestCase(XMLBaseTest):
    """
    Tests for L{DirectoryService}.
    """

    url = "ldap://localhost/"
    baseDN = u"dc=calendarserver,dc=org"
    realmName = unicode(url)


    def setUp(self):
        if MockLdap is None:
            raise SkipTest("No MockLdap available")

        def matches(self, dn, attrs, upcall=MockLDAPFilterTest.matches):
            if upcall(self, dn, attrs):
                return True
            else:
                return mockldap_matches(self, dn, attrs)


        self.patch(MockLDAPFilterTest, "_parse_expression", mockldap_parse)
        self.patch(MockLDAPFilterTest, "matches", mockldap_matches)

        self.xmlSeedService = xmlService(self.mktemp())
        self.mockData = mockDirectoryDataFromXMLService(self.xmlSeedService)

        if False:
            from pprint import pprint
            print("")
            print("-" * 80)
            pprint(self.mockData)
            print("-" * 80)

        self.mockLDAP = MockLdap(self.mockData)
        self.mockLDAP.start()


    def tearDown(self):
        self.mockLDAP.stop()


    def service(self, **kwargs):
        svc = TestService(
            url=self.url,
            baseDN=self.baseDN,
            fieldNameToAttributesMap=TEST_FIELDNAME_MAP,
            recordTypeSchemas=MappingProxyType({
                RecordType.user: RecordTypeSchema(
                    relativeDN=u"cn=user",

                    # (objectClass=inetOrgPerson)
                    attributes=(
                        (
                            LDAPAttribute.objectClass.value,
                            LDAPObjectClass.inetOrgPerson.value,
                        ),
                    ),
                ),

                RecordType.group: RecordTypeSchema(
                    relativeDN=u"cn=group",

                    # (objectClass=groupOfNames)
                    attributes=(
                        (
                            LDAPAttribute.objectClass.value,
                            LDAPObjectClass.groupOfUniqueNames.value,
                        ),
                    ),
                ),
            }),
            **kwargs
        )
        svc.fieldName = ConstantsContainer(
            (svc.fieldName, TestFieldName)
        )
        return svc



class DirectoryServiceQueryTestMixIn(BaseDirectoryServiceQueryTestMixIn):
    def test_queryNot(self):
        return BaseDirectoryServiceQueryTestMixIn.test_queryNot(self)
    test_queryNot.todo = "?"


    def test_queryNotNoIndex(self):
        return BaseDirectoryServiceQueryTestMixIn.test_queryNotNoIndex(self)
    test_queryNotNoIndex.todo = "?"


    def test_queryStartsWithNot(self):
        return BaseDirectoryServiceQueryTestMixIn.test_queryStartsWithNot(self)
    test_queryStartsWithNot.todo = "?"


    def test_queryStartsWithNotAny(self):
        return (
            BaseDirectoryServiceQueryTestMixIn
            .test_queryStartsWithNotAny(self)
        )
    test_queryStartsWithNotAny.todo = "?"


    def test_queryStartsWithNotNoIndex(self):
        return (
            BaseDirectoryServiceQueryTestMixIn
            .test_queryStartsWithNotNoIndex(self)
        )
    test_queryStartsWithNotNoIndex.todo = "?"


    def test_queryContainsNot(self):
        return BaseDirectoryServiceQueryTestMixIn.test_queryContainsNot(self)
    test_queryContainsNot.todo = "?"


    def test_queryContainsNotNoIndex(self):
        return (
            BaseDirectoryServiceQueryTestMixIn
            .test_queryContainsNotNoIndex(self)
        )
    test_queryContainsNotNoIndex.todo = "?"



class DirectoryServiceMutableTestMixIn(BaseDirectoryServiceMutableTestMixIn):
    def test_removeRecord(self):
        return BaseDirectoryServiceMutableTestMixIn.test_removeRecord(self)
    test_removeRecord.todo = "?"


    def test_removeRecordNoExist(self):
        return (
            BaseDirectoryServiceMutableTestMixIn.test_removeRecordNoExist(self)
        )
    test_removeRecordNoExist.todo = "?"


    def test_addRecord(self):
        return BaseDirectoryServiceMutableTestMixIn.test_addRecord(self)
    test_addRecord.todo = "?"


    def test_updateRecord(self):
        return BaseDirectoryServiceMutableTestMixIn.test_updateRecord(self)
    test_updateRecord.todo = "?"


    def test_addRecordNoCreate(self):
        raise NotImplementedError()
        return (
            BaseDirectoryServiceMutableTestMixIn.test_addRecordNoCreate(self)
        )
    test_addRecordNoCreate.todo = "?"



class DirectoryServiceConnectionTestMixIn(object):
    @inlineCallbacks
    def test_connect_defaults(self):
        """
        Connect with default arguments.
        """
        service = self.service()
        connection = yield service._connect()

        self.assertEquals(connection.methods_called(), ["initialize"])

        for option in (
            ldap.OPT_TIMEOUT,
            ldap.OPT_X_TLS_CACERTFILE,
            ldap.OPT_X_TLS_CACERTDIR,
            ldap.OPT_DEBUG_LEVEL,
        ):
            self.assertRaises(KeyError, connection.get_option, option)

        self.assertFalse(connection.tls_enabled)


    @inlineCallbacks
    def test_connect_withUsernamePassword_invalid(self):
        """
        Connect with UsernamePassword credentials.
        """
        credentials = UsernamePassword(
            u"uid=wsanchez,cn=user,{0}".format(self.baseDN),
            u"__password__"
        )
        service = self.service(credentials=credentials)
        try:
            yield service._connect()
        except LDAPBindAuthError:
            pass
        else:
            self.fail("Should have raised LDAPBindAuthError")


    @inlineCallbacks
    def test_connect_withUsernamePassword_valid(self):
        """
        Connect with UsernamePassword credentials.
        """
        credentials = UsernamePassword(
            u"uid=wsanchez,cn=user,{0}".format(self.baseDN),
            u"zehcnasw"
        )
        service = self.service(credentials=credentials)
        connection = yield service._connect()

        self.assertEquals(
            connection.methods_called(),
            ["initialize", "simple_bind_s"]
        )


    @inlineCallbacks
    def test_connect_withOptions(self):
        """
        Connect with default arguments.
        """
        service = self.service(
            timeout=18,
            tlsCACertificateFile=FilePath("/path/to/cert"),
            tlsCACertificateDirectory=FilePath("/path/to/certdir"),
            _debug=True,
        )
        connection = yield service._connect()

        self.assertEquals(
            connection.methods_called(),
            [
                "initialize",
                "set_option", "set_option", "set_option", "set_option",
            ]
        )

        opt = lambda k: connection.get_option(k)

        self.assertEquals(opt(ldap.OPT_TIMEOUT), 18)
        self.assertEquals(opt(ldap.OPT_X_TLS_CACERTFILE), "/path/to/cert")
        self.assertEquals(opt(ldap.OPT_X_TLS_CACERTDIR), "/path/to/certdir")
        self.assertEquals(opt(ldap.OPT_DEBUG_LEVEL), 255)

        # Tested in test_connect_defaults, but test again here since we're
        # setting SSL options and we want to be sure they don't somehow enable
        # SSL implicitly.
        self.assertFalse(connection.tls_enabled)


    @inlineCallbacks
    def test_connect_withTLS(self):
        """
        Connect with TLS enabled.
        """
        service = self.service(useTLS=True)
        connection = yield service._connect()

        self.assertEquals(
            connection.methods_called(),
            ["initialize", "start_tls_s"]
        )

        self.assertTrue(connection.tls_enabled)



class DirectoryServiceTest(
    BaseTestCase,
    DirectoryServiceConvenienceTestMixIn,
    DirectoryServiceRealmTestMixIn,
    DirectoryServiceQueryTestMixIn,
    DirectoryServiceMutableTestMixIn,
    DirectoryServiceConnectionTestMixIn,
    test_directory.BaseDirectoryServiceTest,
    unittest.TestCase,
):
    serviceClass = DirectoryService
    directoryRecordClass = DirectoryRecord


    def test_repr(self):
        service = self.service()

        self.assertEquals(repr(service), u"<TestService u'ldap://localhost/'>")


class ExtraFiltersTest(BaseTestCase, unittest.TestCase):

    def test_extraFilters(self):
        extraFilters = {
            RecordType.user: "(foo=1)",
        }
        service = self.service(extraFilters=extraFilters)
        self.assertEquals(
            "(&(foo=1)(bar=2))",
            service._addExtraFilter(RecordType.user, "(bar=2)")
        )
        self.assertEquals(
            "(bar=2)",
            service._addExtraFilter(RecordType.group, "(bar=2)")
        )

        service = self.service(extraFilters=None)
        self.assertEquals(
            "(bar=2)",
            service._addExtraFilter(RecordType.group, "(bar=2)")
        )

class RecordsFromReplyTest(BaseTestCase, unittest.TestCase):

    def test_boolean(self):
        service = self.service()
        reply = (
            (
                "dn",
                {
                    "__who_uid__": u"active",
                    "foo": u"active",
                    "bar": u"true",
                    "unknown": u"unknown",  # will ignore unknown LDAP attrs
                }
            ),
            (
                "dn",
                {
                    "__who_uid__": u"inactive",
                    "foo": u"inactive",
                    "bar": u"false",
                }
            ),
        )
        records = service._recordsFromReply(reply, recordType=RecordType.user)

        self.assertTrue(records[0].boolean1)   # foo == active so True
        self.assertFalse(records[0].boolean2)  # foo != true so False
        self.assertTrue(records[0].boolean3)  # bar == true so True

        self.assertFalse(records[1].boolean1)  # foo != active so False
        self.assertFalse(records[1].boolean2)  # foo != true so False
        self.assertFalse(records[1].boolean3)  # bar != true so False


    def test_multipleChoice(self):
        service = self.service()
        reply = (
            (
                "dn",
                {
                    "__who_uid__": u"two",
                    "testField": u"Two",
                }
            ),
            (
                "dn",
                {
                    "__who_uid__": u"four",
                    "testField": u"Four",
                }
            ),
        )
        records = service._recordsFromReply(reply, recordType=RecordType.user)
        self.assertEquals(records[0].multiChoice, TestFieldWithChoices.two)

        # "Four" is not a valid value, so it won't get set
        self.assertFalse(service.fieldName.multiChoice in records[1].fields)


    def test_multipleAttributes(self):
        """
        Multiple LDAP attributes can be the source for a single record field.
        If it's a single-value field, the *first* attribute in the map that
        is in the results will be used for the field value.  If it's a multi
        value field, then the value will be a list extended from the values
        in each associated LDAP attribute, in the order of the map entry.

        For example, in the test map above, the uid field is mapped to __who_uid__
        and altuid in that order.  When the LDAP results are parsed, the __who_uid__
        attribute is checked first, and if it has a value it is used; otherwise
        the value from altuid would be used.

        In the case of "fullNames" which is multi-value, because the order in the test
        map is "altname" then "cn", when the LDAP results are parsed, fullNames
        will end up being set to a list comprising of first the altname values
        then the cn values.
        """
        service = self.service()
        reply = (
            (
                "dn",
                {
                    "__who_uid__": u"zero",
                    "altuid": u"altzero",
                    "cn": [u"cn-name", "another-cn"],
                    "altname": [u"alt-name", "another-alt"],
                }
            ),
            (
                "dn",
                {
                    "altuid": u"one",
                    "altname": [u"alt-name"],
                    "cn": [u"cn-name"],
                }
            ),
            (
                "dn",
                {
                    "__who_uid__": u"two",
                    "cn": [u"cn-name"],
                }
            ),
        )
        records = service._recordsFromReply(reply, recordType=RecordType.user)

        self.assertEquals(records[0].uid, "zero")
        self.assertEquals(records[0].fullNames, ["alt-name", "another-alt", "cn-name", "another-cn"])
        self.assertEquals(records[1].uid, "one")
        self.assertEquals(records[1].fullNames, ["alt-name", "cn-name"])
        self.assertEquals(records[2].uid, "two")
        self.assertEquals(records[2].fullNames, ["cn-name"])


def mockDirectoryDataFromXMLService(service):
    dc0 = u"org"
    dc1 = u"calendarserver"

    data = {
        u"dc={dc0}".format(dc0=dc0): dict(dc=dc0),
        u"dc={dc1},dc={dc0}".format(dc1=dc1, dc0=dc0): dict(dc=[dc1, dc0]),
        u"cn=user,dc={dc1},dc={dc0}".format(dc1=dc1, dc0=dc0): dict(dc=[dc1, dc0]),
        u"cn=group,dc={dc1},dc={dc0}".format(dc1=dc1, dc0=dc0): dict(dc=[dc1, dc0]),
    }

    def toUnicode(obj):
        if isinstance(obj, (NamedConstant, ValueConstant)):
            return obj.name

        if isinstance(obj, (tuple, list)):
            return [unicode(x) for x in obj]

        return unicode(obj)


    def tuplify(record, fieldName):
        fieldValue = record.fields[fieldName]

        if fieldName is BaseFieldName.recordType:
            schema = DEFAULT_RECORDTYPE_SCHEMAS[fieldValue]

            return schema.attributes

        else:

            # mockldap requires data in this structure: {dn: {attr: [values]}}
            if not isinstance(fieldValue, (tuple, list)):
                fieldValue = [fieldValue]

            # Question: why convert to unicode?  We get utf-8 encoded strings
            # back from a real LDAP server, don't we?
            value = toUnicode(fieldValue)

            return (
                (name, value)
                for name in TEST_FIELDNAME_MAP.get(
                    fieldName, (u"__" + fieldName.name,)
                )
            )

    for records in service.index[service.fieldName.uid].itervalues():
        for record in records:
            dn = u"uid={uid},cn={cn},dc={dc1},dc={dc0}".format(
                uid=record.shortNames[0],
                cn=record.recordType.name,
                dc1=dc1, dc0=dc0
            )

            recordData = dict(chain(*(
                list(tuplify(record, fieldName))
                for fieldName in service.fieldName.iterconstants()
                if fieldName in record.fields
            )))

            data[dn] = recordData

    return data



#
# mockldap patches
# See https://bitbucket.org/psagers/mockldap/issue/4/
#

class WildcardExpression(object):
    first = None
    middle = []
    last = None



def mockldap_parse(self):
    match = self.TEST_RE.match(self.content)

    if match is None:
        raise ldap.FILTER_ERROR(
            u"Failed to parse filter item %r at pos %d"
            % (self.content, self.start)
        )

    self.attr, self.op, valueExpression = match.groups()

    if self.op != "=":
        raise MockLDAPUnsupportedOp(
            u"Operation %r is not supported" % (self.op,)
        )


    def unescape(value):
        return self.UNESCAPE_RE.sub(lambda m: chr(int(m.group(1), 16)), value)

    if (u"*" in valueExpression):
        # Wild card expression

        values = [unescape(value) for value in valueExpression.split(u"*")]

        exp = WildcardExpression()

        if not valueExpression.startswith(u"*"):
            exp.first = values.pop(0)

        if not valueExpression.endswith(u"*"):
            exp.last = values.pop(-1)

        exp.middle = values

        self.value = exp

    else:
        self.value = unescape(valueExpression)



def mockldap_matches(self, dn, attrs):
    values = attrs.get(self.attr)

    if values is None:
        return False

    if type(values) is unicode:
        values = [values]

    # Case insensitive?  Always true in LDAP, it seems.
    if True:
        normalize = lambda s: s.lower()
    else:
        normalize = lambda s: s

    if isinstance(self.value, WildcardExpression):
        def match_substrings_in_order(substrings, value, start, end):
            for substring in substrings:
                if not substring:
                    continue

                i = value.find(substring, start, end)
                if i == -1:
                    # Match fails for this substring
                    return False

                # Move start up past this substring substring before testing
                # the next substring
                start = i + len(substring)

            # No mismatches
            return True

        for value in values:
            value = normalize(value)

            start = 0
            end = len(value)

            if self.value.first is not None:
                if not value.startswith(normalize(self.value.first)):
                    continue
                start = len(self.value.first)

            if self.value.last is not None:
                if not value[start:].endswith(normalize(self.value.last)):
                    continue
                end -= len(self.value.last)

            if self.value.middle:
                if not match_substrings_in_order(
                    (normalize(s) for s in self.value.middle),
                    value, start, end
                ):
                    continue

            return True

        return False

    return normalize(self.value) in (normalize(s) for s in values)



##
# Copyright (c) 2013-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Aggregate directory service tests.
"""

from twisted.python.components import proxyForInterface
from twisted.trial import unittest

from ..idirectory import IDirectoryService, DirectoryConfigurationError
from ..directory import DirectoryRecord
from ..aggregate import DirectoryService
from ..util import ConstantsContainer
from . import test_directory, test_xml
from .test_xml import (
    BaseTest as XMLBaseTest,
    QueryMixIn, xmlService, TestService as XMLTestService,
    DirectoryServiceConvenienceTestMixIn,
)


class BaseTest(XMLBaseTest):
    def service(self, services=None):
        if services is None:
            services = (self.xmlService(),)

        #
        # Make sure aggregate DirectoryService isn't making
        # implementation assumptions about the IDirectoryService
        # objects it gets.
        #
        services = tuple((
            proxyForInterface(IDirectoryService)(s)
            for s in services
        ))

        class TestService(DirectoryService, QueryMixIn):
            pass

        return TestService(u"xyzzy", services)


    def xmlService(self, xmlData=None, serviceClass=None):
        return xmlService(
            self.mktemp(),
            xmlData=xmlData,
            serviceClass=serviceClass
        )



class DirectoryServiceTest(
    unittest.TestCase,
    BaseTest,
    DirectoryServiceConvenienceTestMixIn,
    test_directory.BaseDirectoryServiceTest
):
    serviceClass = DirectoryService
    directoryRecordClass = DirectoryRecord

    def test_repr(self):
        service = self.service()
        self.assertEquals(repr(service), "<TestService u'xyzzy'>")



class DirectoryServiceQueryTest(
    unittest.TestCase,
    BaseTest,
    test_xml.DirectoryServiceQueryTestMixIn
):
    pass



class DirectoryServiceImmutableTest(
    BaseTest,
    test_directory.BaseDirectoryServiceImmutableTest,
):
    serviceClass = DirectoryService
    directoryRecordClass = DirectoryRecord



class AggregatedBaseTest(BaseTest):
    def service(self):
        class UsersDirectoryService(XMLTestService):
            recordType = ConstantsContainer((XMLTestService.recordType.user,))

        class GroupsDirectoryService(XMLTestService):
            recordType = ConstantsContainer((XMLTestService.recordType.group,))

        usersService = self.xmlService(
            xmlData=testXMLConfigUsers,
            serviceClass=UsersDirectoryService
        )
        groupsService = self.xmlService(
            xmlData=testXMLConfigGroups,
            serviceClass=GroupsDirectoryService
        )

        return BaseTest.service(
            self,
            services=(usersService, groupsService)
        )



class DirectoryServiceAggregatedBaseTest(
    AggregatedBaseTest,
    DirectoryServiceTest,
):
    pass



class DirectoryServiceAggregatedQueryTest(
    unittest.TestCase,
    AggregatedBaseTest,
    test_xml.DirectoryServiceQueryTestMixIn,
):
    pass



class DirectoryServiceAggregatedImmutableTest(
    AggregatedBaseTest,
    test_directory.BaseDirectoryServiceImmutableTest,
):
    pass



class DirectoryServiceTests(BaseTest, unittest.TestCase):
    def test_conflictingRecordTypes(self):
        self.assertRaises(
            DirectoryConfigurationError,
            self.service,
            services=(self.xmlService(), self.xmlService(testXMLConfigUsers)),
        )



testXMLConfigUsers = """<?xml version="1.0" encoding="utf-8"?>

<directory realm="Users Realm">

  <record type="user">
    <uid>__wsanchez__</uid>
    <short-name>wsanchez</short-name>
    <short-name>wilfredo_sanchez</short-name>
    <full-name>Wilfredo Sanchez</full-name>
    <password>zehcnasw</password>
    <email>wsanchez@bitbucket.calendarserver.org</email>
    <email>wsanchez@devnull.twistedmatrix.com</email>
  </record>

  <record type="user">
    <uid>__glyph__</uid>
    <short-name>glyph</short-name>
    <full-name>Glyph Lefkowitz</full-name>
    <password>hpylg</password>
    <email>glyph@bitbucket.calendarserver.org</email>
    <email>glyph@devnull.twistedmatrix.com</email>
  </record>

  <record type="user">
    <uid>__sagen__</uid>
    <short-name>sagen</short-name>
    <full-name>Morgen Sagen</full-name>
    <password>negas</password>
    <email>sagen@bitbucket.calendarserver.org</email>
    <email>shared@example.com</email>
  </record>

  <record type="user">
    <uid>__cdaboo__</uid>
    <short-name>cdaboo</short-name>
    <full-name>Cyrus Daboo</full-name>
    <password>suryc</password>
    <email>cdaboo@bitbucket.calendarserver.org</email>
  </record>

  <record type="user">
    <uid>__dre__</uid>
    <guid>A3B1158F-0564-4F5B-81E4-A89EA5FF81B0</guid>
    <short-name>dre</short-name>
    <full-name>Andre LaBranche</full-name>
    <password>erd</password>
    <email>dre@bitbucket.calendarserver.org</email>
    <email>shared@example.com</email>
  </record>

  <record type="user">
    <uid>__exarkun__</uid>
    <short-name>exarkun</short-name>
    <full-name>Jean-Paul Calderone</full-name>
    <password>nucraxe</password>
    <email>exarkun@devnull.twistedmatrix.com</email>
  </record>

  <record type="user">
    <uid>__dreid__</uid>
    <short-name>dreid</short-name>
    <full-name>David Reid</full-name>
    <password>dierd</password>
    <email>dreid@devnull.twistedmatrix.com</email>
  </record>

  <record> <!-- type defaults to "user" -->
    <uid>__joe__</uid>
    <short-name>joe</short-name>
    <full-name>Joe Schmoe</full-name>
    <password>eoj</password>
    <email>joe@example.com</email>
  </record>

  <record> <!-- type defaults to "user" -->
    <uid>__alyssa__</uid>
    <short-name>alyssa</short-name>
    <full-name>Alyssa P. Hacker</full-name>
    <password>assyla</password>
    <email>alyssa@example.com</email>
  </record>

</directory>
"""


testXMLConfigGroups = """<?xml version="1.0" encoding="utf-8"?>

<directory realm="Groups Realm">

  <record type="group">
    <uid>__calendar-dev__</uid>
    <short-name>calendar-dev</short-name>
    <full-name>Calendar Server developers</full-name>
    <email>dev@bitbucket.calendarserver.org</email>
    <member-uid>__wsanchez__</member-uid>
    <member-uid>__glyph__</member-uid>
    <member-uid>__sagen__</member-uid>
    <member-uid>__cdaboo__</member-uid>
    <member-uid>__dre__</member-uid>
  </record>

  <record type="group">
    <uid>__twisted__</uid>
    <short-name>twisted</short-name>
    <full-name>Twisted Matrix Laboratories</full-name>
    <email>hack@devnull.twistedmatrix.com</email>
    <member-uid>__wsanchez__</member-uid>
    <member-uid>__glyph__</member-uid>
    <member-uid>__exarkun__</member-uid>
    <member-uid>__dreid__</member-uid>
    <member-uid>__dre__</member-uid>
  </record>

  <record type="group">
    <uid>__developers__</uid>
    <short-name>developers</short-name>
    <full-name>All Developers</full-name>
    <member-uid>__calendar-dev__</member-uid>
    <member-uid>__twisted__</member-uid>
    <member-uid>__alyssa__</member-uid>
  </record>

</directory>
"""



##
# Copyright (c) 2013-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
XML directory service tests.
"""

from time import sleep
from uuid import UUID
from textwrap import dedent

from twisted.trial import unittest
from twisted.python.constants import Names, NamedConstant
from twisted.python.filepath import FilePath
from twisted.internet.defer import inlineCallbacks

from ..idirectory import DirectoryAvailabilityError, NoSuchRecordError
from ..expression import (
    CompoundExpression, Operand,
    MatchExpression, MatchType, MatchFlags
)
from ..xml import ParseError, DirectoryService, DirectoryRecord
from . import test_index



class BaseTest(object):
    def service(self, subClass=None, xmlData=None):
        return xmlService(
            self.mktemp(),
            xmlData=xmlData,
            serviceClass=subClass
        )


    def assertRecords(self, records, uids):
        self.assertEquals(
            frozenset((record.uid for record in records)),
            frozenset((uids)),
        )



class DirectoryServiceConvenienceTestMixIn(object):
    @inlineCallbacks
    def test_recordWithUID(self):
        service = self.service()

        record = (yield service.recordWithUID(u"__null__"))
        self.assertEquals(record, None)

        record = (yield service.recordWithUID(u"__wsanchez__"))
        self.assertEquals(record.uid, u"__wsanchez__")


    @inlineCallbacks
    def test_recordWithGUID(self):
        service = self.service()

        record = (
            yield service.recordWithGUID(
                UUID("6C495FCD-7E78-4D5C-AA66-BC890AD04C9D")
            )
        )
        self.assertEquals(record, None)

        record = (
            yield service.recordWithGUID(
                UUID("A3B1158F-0564-4F5B-81E4-A89EA5FF81B0")
            )
        )
        self.assertEquals(record.uid, u"__dre__")


    @inlineCallbacks
    def test_recordsWithRecordType_unknown(self):
        service = self.service()

        records = (
            yield service.recordsWithRecordType(UnknownConstant.unknown)
        )
        self.assertEquals(set(records), set())


    @inlineCallbacks
    def test_recordsWithRecordType(self):
        service = self.service()

        records = (
            yield service.recordsWithRecordType(service.recordType.user)
        )
        self.assertRecords(
            records,
            (
                u"__wsanchez__",
                u"__glyph__",
                u"__sagen__",
                u"__cdaboo__",
                u"__dre__",
                u"__exarkun__",
                u"__dreid__",
                u"__alyssa__",
                u"__joe__",
            ),
        )

        records = (
            yield service.recordsWithRecordType(service.recordType.group)
        )
        self.assertRecords(
            records,
            (
                u"__calendar-dev__",
                u"__twisted__",
                u"__developers__",
            ),
        )


    @inlineCallbacks
    def test_recordWithShortName(self):
        service = self.service()

        record = (
            yield service.recordWithShortName(
                service.recordType.user,
                u"null",
            )
        )
        self.assertEquals(record, None)

        record = (
            yield service.recordWithShortName(
                service.recordType.user,
                u"wsanchez",
            )
        )
        self.assertEquals(record.uid, u"__wsanchez__")

        record = (
            yield service.recordWithShortName(
                service.recordType.user,
                u"wilfredo_sanchez",
            )
        )
        self.assertEquals(record.uid, u"__wsanchez__")


    @inlineCallbacks
    def test_recordsWithEmailAddress(self):
        service = self.service()

        records = (
            yield service.recordsWithEmailAddress(
                u"wsanchez@bitbucket.calendarserver.org"
            )
        )
        self.assertRecords(records, (u"__wsanchez__",))

        records = (
            yield service.recordsWithEmailAddress(
                u"wsanchez@devnull.twistedmatrix.com"
            )
        )
        self.assertRecords(records, (u"__wsanchez__",))

        records = (
            yield service.recordsWithEmailAddress(
                u"shared@example.com"
            )
        )
        self.assertRecords(records, (u"__sagen__", u"__dre__"))


    @inlineCallbacks
    def test_limitResults(self):
        """
        Make sure limitResults does limit results.
        """

        service = self.service()

        records = (
            yield service.recordsWithRecordType(
                service.recordType.user,
                limitResults=3
            )
        )
        self.assertEquals(len(records), 3)

        records = (
            yield service.recordsWithRecordType(
                service.recordType.user,
                limitResults=1000
            )
        )
        self.assertEquals(len(records), 9)



class DirectoryServiceRealmTestMixIn(object):
    def test_realmNameImmutable(self):
        def setRealmName():
            service = self.service()
            service.realmName = u"foo"

        self.assertRaises(AttributeError, setRealmName)



class DirectoryServiceQueryTestMixIn(object):
    @inlineCallbacks
    def test_queryAnd(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            CompoundExpression(
                (
                    service.query(u"emailAddresses", u"shared@example.com"),
                    service.query(u"shortNames", u"sagen"),
                ),
                operand=Operand.AND
            )
        )
        self.assertRecords(records, (u"__sagen__",))


    @inlineCallbacks
    def test_queryAndNoneFirst(self):
        """
        Test optimized case, where first expression yields no results.
        """
        service = self.service()
        records = yield service.recordsFromExpression(
            CompoundExpression(
                (
                    service.query(u"emailAddresses", u"nobody@example.com"),
                    service.query(u"shortNames", u"sagen"),
                ),
                operand=Operand.AND
            )
        )
        self.assertRecords(records, ())


    @inlineCallbacks
    def test_queryOr(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            CompoundExpression(
                (
                    service.query(u"emailAddresses", u"shared@example.com"),
                    service.query(u"shortNames", u"wsanchez"),
                ),
                operand=Operand.OR
            )
        )
        self.assertRecords(
            records,
            (u"__sagen__", u"__dre__", u"__wsanchez__")
        )


    @inlineCallbacks
    def test_queryNot(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            CompoundExpression(
                (
                    service.query(
                        u"emailAddresses", u"shared@example.com"
                    ),
                    service.query(
                        u"shortNames", u"sagen",
                        flags=MatchFlags.NOT
                    ),
                ),
                operand=Operand.AND
            )
        )
        self.assertRecords(records, (u"__dre__",))


    @inlineCallbacks
    def test_queryNotNoIndex(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            CompoundExpression(
                (
                    service.query(u"emailAddresses", u"shared@example.com"),
                    service.query(
                        u"fullNames", u"Andre LaBranche",
                        flags=MatchFlags.NOT
                    ),
                ),
                operand=Operand.AND
            )
        )
        self.assertRecords(records, (u"__sagen__",))


    @inlineCallbacks
    def test_queryCaseInsensitive(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"shortNames", u"SagEn",
                flags=MatchFlags.caseInsensitive
            )
        )
        self.assertRecords(records, (u"__sagen__",))


    @inlineCallbacks
    def test_queryCaseInsensitiveNoIndex(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"fullNames", u"moRGen SAGen",
                flags=MatchFlags.caseInsensitive
            )
        )
        self.assertRecords(records, (u"__sagen__",))


    @inlineCallbacks
    def test_queryStartsWith(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"shortNames", u"wil",
                matchType=MatchType.startsWith
            )
        )
        self.assertRecords(records, (u"__wsanchez__",))


    @inlineCallbacks
    def test_queryStartsWithNoIndex(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"fullNames", u"Wilfredo",
                matchType=MatchType.startsWith
            )
        )
        self.assertRecords(records, (u"__wsanchez__",))


    @inlineCallbacks
    def test_queryStartsWithNot(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"shortNames", u"w",
                matchType=MatchType.startsWith,
                flags=MatchFlags.NOT,
            )
        )
        self.assertRecords(
            records,
            (
                u"__alyssa__",
                u"__calendar-dev__",
                u"__cdaboo__",
                u"__developers__",
                u"__dre__",
                u"__dreid__",
                u"__exarkun__",
                u"__glyph__",
                u"__joe__",
                u"__sagen__",
                u"__twisted__",
            ),
        )


    @inlineCallbacks
    def test_queryStartsWithNotAny(self):
        """
        FIXME?: In the this case, the record __wsanchez__ has two
        shortNames, and one doesn't match the query.  Should it be
        included or not?  It is, because one matches the query, but
        should NOT require that all match?
        """
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"shortNames", u"wil",
                matchType=MatchType.startsWith,
                flags=MatchFlags.NOT,
            )
        )
        self.assertRecords(
            records,
            (
                u"__alyssa__",
                u"__calendar-dev__",
                u"__cdaboo__",
                u"__developers__",
                u"__dre__",
                u"__dreid__",
                u"__exarkun__",
                u"__glyph__",
                u"__joe__",
                u"__sagen__",
                u"__twisted__",
                u"__wsanchez__",
            ),
        )


    @inlineCallbacks
    def test_queryStartsWithNotNoIndex(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"fullNames", u"Andre",
                matchType=MatchType.startsWith,
                flags=MatchFlags.NOT,
            )
        )
        self.assertRecords(
            records,
            (
                u"__alyssa__",
                u"__calendar-dev__",
                u"__cdaboo__",
                u"__developers__",
                u"__dreid__",
                u"__exarkun__",
                u"__glyph__",
                u"__joe__",
                u"__sagen__",
                u"__twisted__",
                u"__wsanchez__",
            ),
        )


    @inlineCallbacks
    def test_queryStartsWithCaseInsensitive(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"shortNames", u"WIL",
                matchType=MatchType.startsWith,
                flags=MatchFlags.caseInsensitive,
            )
        )
        self.assertRecords(records, (u"__wsanchez__",))


    @inlineCallbacks
    def test_queryStartsWithCaseInsensitiveNoIndex(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"fullNames", u"wilfrEdo",
                matchType=MatchType.startsWith,
                flags=MatchFlags.caseInsensitive,
            )
        )
        self.assertRecords(records, (u"__wsanchez__",))


    @inlineCallbacks
    def test_queryContains(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"shortNames", u"sanchez",
                matchType=MatchType.contains,
            )
        )
        self.assertRecords(records, (u"__wsanchez__",))


    @inlineCallbacks
    def test_queryContainsNoIndex(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"fullNames", u"fred",
                matchType=MatchType.contains,
            )
        )
        self.assertRecords(records, (u"__wsanchez__",))


    @inlineCallbacks
    def test_queryContainsNot(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"shortNames", u"sanchez",
                matchType=MatchType.contains,
                flags=MatchFlags.NOT,
            )
        )
        self.assertRecords(
            records,
            (
                u"__alyssa__",
                u"__calendar-dev__",
                u"__cdaboo__",
                u"__developers__",
                u"__dre__",
                u"__dreid__",
                u"__exarkun__",
                u"__glyph__",
                u"__joe__",
                u"__sagen__",
                u"__twisted__",
            ),
        )


    @inlineCallbacks
    def test_queryContainsNotNoIndex(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"fullNames", u"ranch",
                matchType=MatchType.contains,
                flags=MatchFlags.NOT,
            )
        )
        self.assertRecords(
            records,
            (
                u"__alyssa__",
                u"__calendar-dev__",
                u"__cdaboo__",
                u"__developers__",
                u"__dreid__",
                u"__exarkun__",
                u"__glyph__",
                u"__joe__",
                u"__sagen__",
                u"__twisted__",
                u"__wsanchez__",
            ),
        )


    @inlineCallbacks
    def test_queryContainsCaseInsensitive(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"shortNames", u"Sanchez",
                matchType=MatchType.contains,
                flags=MatchFlags.caseInsensitive,
            )
        )
        self.assertRecords(records, (u"__wsanchez__",))


    @inlineCallbacks
    def test_queryContainsCaseInsensitiveNoIndex(self):
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"fullNames", u"frEdo",
                matchType=MatchType.contains,
                flags=MatchFlags.caseInsensitive,
            )
        )
        self.assertRecords(records, (u"__wsanchez__",))


    @inlineCallbacks
    def test_queryWithRecordTypes(self):
        """
        Verify that results are limited to the requested recordTypes
        """
        service = self.service()
        records = yield service.recordsFromExpression(
            service.query(
                u"fullNames", u"e",
                matchType=MatchType.contains,
            ),
            recordTypes=(service.recordType.group,)
        )
        # Note: only contains groups; the users that would normally match
        # have been filtered out
        self.assertRecords(
            records,
            (u"__calendar-dev__", u"__developers__", u"__twisted__")
        )



class DirectoryServiceMutableTestMixIn(object):
    @inlineCallbacks
    def test_updateRecord(self):
        service = self.service()

        record = (yield service.recordWithUID(u"__wsanchez__"))

        fields = record.fields.copy()
        fields[service.fieldName.fullNames] = [u"Wilfredo Sanchez Vega"]

        updatedRecord = DirectoryRecord(service, fields)
        yield service.updateRecords((updatedRecord,))

        # Verify change is present immediately
        record = (yield service.recordWithUID(u"__wsanchez__"))
        self.assertEquals(
            set(record.fullNames),
            set((u"Wilfredo Sanchez Vega",))
        )

        # Verify change is persisted
        service.flush()
        record = (yield service.recordWithUID(u"__wsanchez__"))
        self.assertEquals(
            set(record.fullNames),
            set((u"Wilfredo Sanchez Vega",))
        )


    @inlineCallbacks
    def test_addRecord(self):
        service = self.service()

        newRecord = DirectoryRecord(
            service,
            fields={
                service.fieldName.uid: u"__plugh__",
                service.fieldName.recordType: service.recordType.user,
                service.fieldName.shortNames: (u"plugh",),
                service.fieldName.password: u"",
            }
        )

        yield service.updateRecords((newRecord,), create=True)

        # Verify change is present immediately
        record = (yield service.recordWithUID(u"__plugh__"))
        self.assertEquals(set(record.shortNames), set((u"plugh",)))
        self.assertEquals(record.password, u"")

        # Verify change is persisted
        service.flush()
        record = (yield service.recordWithUID(u"__plugh__"))
        self.assertEquals(set(record.shortNames), set((u"plugh",)))


    def test_addRecordNoCreate(self):
        service = self.service()

        newRecord = DirectoryRecord(
            service,
            fields={
                service.fieldName.uid: u"__plugh__",
                service.fieldName.recordType: service.recordType.user,
                service.fieldName.shortNames: (u"plugh",),
            }
        )

        return self.assertFailure(
            service.updateRecords((newRecord,)),
            NoSuchRecordError
        )


    @inlineCallbacks
    def test_removeRecord(self):
        service = self.service()

        yield service.removeRecords((u"__wsanchez__",))

        # Verify change is present immediately
        self.assertEquals((yield service.recordWithUID(u"__wsanchez__")), None)

        # Verify change is persisted
        service.flush()
        self.assertEquals((yield service.recordWithUID(u"__wsanchez__")), None)


    def test_removeRecordNoExist(self):
        service = self.service()

        return service.removeRecords((u"__plugh__",))



class DirectoryServiceTest(
    unittest.TestCase,
    BaseTest,
    DirectoryServiceConvenienceTestMixIn,
    DirectoryServiceRealmTestMixIn,
    DirectoryServiceQueryTestMixIn,
    DirectoryServiceMutableTestMixIn,
    test_index.BaseDirectoryServiceTest,
):
    serviceClass = DirectoryService
    directoryRecordClass = DirectoryRecord

    def test_repr(self):
        service = self.service()

        self.assertEquals(repr(service), u"<TestService (not loaded)>")
        service.loadRecords()
        self.assertEquals(repr(service), u"<TestService u'xyzzy'>")



class DirectoryServiceParsingTest(unittest.TestCase, BaseTest):
    def test_reloadInterval(self):
        service = self.service()

        service.loadRecords(stat=False)
        lastRefresh = service._lastRefresh
        self.assertTrue(service._lastRefresh)

        sleep(1)
        service.loadRecords(stat=False)
        self.assertEquals(lastRefresh, service._lastRefresh)


    def test_reloadStat(self):
        service = self.service()

        service.loadRecords(loadNow=True)
        lastRefresh = service._lastRefresh
        self.assertTrue(service._lastRefresh)

        sleep(1)
        service.loadRecords(loadNow=True)
        self.assertEquals(lastRefresh, service._lastRefresh)


    def test_badXML(self):
        service = self.service(xmlData="Hello")

        self.assertRaises(ParseError, service.loadRecords)


    def test_badRootElement(self):
        service = self.service(xmlData=(dedent(
            b"""
            <?xml version="1.0" encoding="utf-8"?>

            <frobnitz />
            """[1:]
        )))

        self.assertRaises(ParseError, service.loadRecords)
        try:
            service.loadRecords()
        except ParseError as e:
            self.assertTrue(str(e).startswith("Incorrect root element"), e)
        else:
            raise AssertionError("Expected ParseError")


    def test_noRealmName(self):
        service = self.service(xmlData=(dedent(
            b"""
            <?xml version="1.0" encoding="utf-8"?>

            <directory />
            """[1:]
        )))

        self.assertRaises(ParseError, service.loadRecords)
        try:
            service.loadRecords()
        except ParseError as e:
            self.assertTrue(str(e).startswith("No realm name"), e)
        else:
            raise AssertionError("Expected ParseError")


    def test_unknownFieldElementsClean(self):
        service = self.service()
        self.assertEquals(set(service.unknownFieldElements), set())


    def test_unknownFieldElementsDirty(self):
        service = self.service(xmlData=(dedent(
            b"""
            <?xml version="1.0" encoding="utf-8"?>

            <directory realm="Unknown Record Types">
              <record type="user">
                <uid>__wsanchez__</uid>
                <short-name>wsanchez</short-name>
                <political-affiliation>Community and Freedom Party</political-affiliation>
              </record>
            </directory>
            """[1:]
        )))
        self.assertEquals(
            set(service.unknownFieldElements),
            set((u"political-affiliation",))
        )


    def test_unknownRecordTypesClean(self):
        service = self.service()
        self.assertEquals(set(service.unknownRecordTypes), set())


    def test_unknownRecordTypesDirty(self):
        service = self.service(xmlData=(dedent(
            b"""
            <?xml version="1.0" encoding="utf-8"?>

            <directory realm="Unknown Record Types">
              <record type="camera">
                <uid>__d600__</uid>
                <short-name>d600</short-name>
                <full-name>Nikon D600</full-name>
              </record>
            </directory>
            """[1:]
        )))

        self.assertEquals(set(service.unknownRecordTypes), set((u"camera",)))


    def test_unknownFieldElementClean(self):
        service = self.service()
        self.assertEquals(set(service.unknownFieldElements), set())


    @inlineCallbacks
    def test_unknownFieldElementDirty(self):
        service = self.service(xmlData=(dedent(
            b"""
            <?xml version="1.0" encoding="utf-8"?>

            <directory realm="Unknown Record Types">
              <record type="user">
                <uid>__wsanchez__</uid>
                <short-name>wsanchez</short-name>
                <nickname>Larry</nickname>
              </record>
            </directory>
            """[1:]
        )))

        self.assertEquals(
            set(service.unknownFieldElements),
            set((u"nickname",))
        )

        # Make sure that the 'short-name', which was parsed prior to nickname
        # is correct.  That is, make sure that we continue as soon as we
        # notice the unknown field name and don't populate the previous field.

        record = yield service.recordWithUID(u"__wsanchez__")

        self.assertEquals(set(record.shortNames), set((u"wsanchez",)))



class DirectoryRecordTest(
    unittest.TestCase,
    BaseTest,
    test_index.BaseDirectoryRecordTest
):

    serviceClass = DirectoryService
    directoryRecordClass = DirectoryRecord


    @inlineCallbacks
    def test_members_group(self):
        service = self.service()

        record = (yield service.recordWithUID(u"__wsanchez__"))
        members = (yield record.members())
        self.assertEquals(set(members), set())

        record = (yield service.recordWithUID(u"__twisted__"))
        members = (yield record.members())
        self.assertEquals(
            set((member.uid for member in members)),
            set(
                (
                    u"__wsanchez__",
                    u"__glyph__",
                    u"__exarkun__",
                    u"__dreid__",
                    u"__dre__",
                )
            )
        )

        record = (yield service.recordWithUID(u"__developers__"))
        members = (yield record.members())
        self.assertEquals(
            set((member.uid for member in members)),
            set(
                (
                    u"__calendar-dev__",
                    u"__twisted__",
                    u"__alyssa__",
                )
            )
        )


    @inlineCallbacks
    def test_memberships(self):
        service = self.service()

        record = (yield service.recordWithUID(u"__wsanchez__"))
        groups = (yield record.groups())
        self.assertEquals(
            set(group.uid for group in groups),
            set(
                (
                    u"__calendar-dev__",
                    u"__twisted__",
                )
            )
        )



class MissingFileTest(unittest.TestCase):

    def test_missingFile(self):
        service = DirectoryService(FilePath(self.mktemp()))

        return self.assertFailure(
            service.recordWithUID(u"missing"),
            DirectoryAvailabilityError
        )



class QueryMixIn(object):
    def query(
        self,
        field, value,
        matchType=MatchType.equals, flags=MatchFlags.none
    ):
        name = getattr(self.fieldName, field)
        assert name is not None
        return MatchExpression(
            name, value,
            matchType=matchType,
            flags=flags,
        )



class TestService(DirectoryService, QueryMixIn):
    pass



class UnknownConstant(Names):
    unknown = NamedConstant()



def xmlService(tmp, xmlData=None, serviceClass=None):
    if xmlData is None:
        xmlData = testXMLConfig

    if serviceClass is None:
        serviceClass = TestService

    filePath = FilePath(tmp)
    filePath.setContent(xmlData)

    try:
        return serviceClass(filePath)
    except Exception as e:
        raise AssertionError(
            "Unable to instantiate XML service {0}: {1}"
            .format(serviceClass, e)
        )



testXMLConfig = b"""<?xml version="1.0" encoding="utf-8"?>

<directory realm="xyzzy">

  <record type="user">
    <uid>__wsanchez__</uid>
    <short-name>wsanchez</short-name>
    <short-name>wilfredo_sanchez</short-name>
    <full-name>Wilfredo Sanchez</full-name>
    <full-name>Wilfredo S\xc3\xa1nchez Vega</full-name>
    <full-name>\xe3\x82\xb5\xe3\x83\xb3\xe3\x83\x81\xe3\x82\xa7\xe3\x82\xba\xe3\x80\x80\xe3\x83\xb4\xe3\x82\xa7\xe3\x82\xac\xe3\x80\x80\xe3\x82\xa6\xe3\x82\xa3\xe3\x83\xab\xe3\x83\x95\xe3\x83\xac\xe3\x83\xbc\xe3\x83\x89</full-name>
<!--
    <given-name>Wilfredo</given-name>
    <given-name>\xe3\x82\xa6\xe3\x82\xa3\xe3\x83\xab\xe3\x83\x95\xe3\x83\xac\xe3\x83\xbc\xe3\x83\x89</given-name>
    <surname>Sanchez</surname>
    <surname>S\xc3\xa1nchez Vega</surname>
    <surname>\xe3\x82\xb5\xe3\x83\xb3\xe3\x83\x81\xe3\x82\xa7\xe3\x82\xba\xe3\x80\x80\xe3\x83\xb4\xe3\x82\xa7\xe3\x82\xac</surname>
 -->
    <password>zehcnasw</password>
    <email>wsanchez@bitbucket.calendarserver.org</email>
    <email>wsanchez@devnull.twistedmatrix.com</email>
  </record>

  <record type="user">
    <uid>__glyph__</uid>
    <short-name>glyph</short-name>
    <full-name>Glyph Lefkowitz</full-name>
    <password>hpylg</password>
    <email>glyph@bitbucket.calendarserver.org</email>
    <email>glyph@devnull.twistedmatrix.com</email>
  </record>

  <record type="user">
    <uid>__sagen__</uid>
    <short-name>sagen</short-name>
    <full-name>Morgen Sagen</full-name>
    <password>negas</password>
    <email>sagen@bitbucket.calendarserver.org</email>
    <email>shared@example.com</email>
  </record>

  <record type="user">
    <uid>__cdaboo__</uid>
    <short-name>cdaboo</short-name>
    <full-name>Cyrus Daboo</full-name>
    <password>suryc</password>
    <email>cdaboo@bitbucket.calendarserver.org</email>
  </record>

  <record type="user">
    <uid>__dre__</uid>
    <guid>A3B1158F-0564-4F5B-81E4-A89EA5FF81B0</guid>
    <short-name>dre</short-name>
    <full-name>Andre LaBranche</full-name>
    <password>erd</password>
    <email>dre@bitbucket.calendarserver.org</email>
    <email>shared@example.com</email>
  </record>

  <record type="user">
    <uid>__exarkun__</uid>
    <short-name>exarkun</short-name>
    <full-name>Jean-Paul Calderone</full-name>
    <password>nucraxe</password>
    <email>exarkun@devnull.twistedmatrix.com</email>
  </record>

  <record type="user">
    <uid>__dreid__</uid>
    <short-name>dreid</short-name>
    <full-name>David Reid</full-name>
    <password>dierd</password>
    <email>dreid@devnull.twistedmatrix.com</email>
  </record>

  <record> <!-- type defaults to "user" -->
    <uid>__joe__</uid>
    <short-name>joe</short-name>
    <full-name>Joe Schmoe</full-name>
    <password>eoj</password>
    <email>joe@example.com</email>
  </record>

  <record> <!-- type defaults to "user" -->
    <uid>__alyssa__</uid>
    <short-name>alyssa</short-name>
    <full-name>Alyssa P. Hacker</full-name>
    <password>assyla</password>
    <email>alyssa@example.com</email>
  </record>

  <record type="group">
    <uid>__calendar-dev__</uid>
    <short-name>calendar-dev</short-name>
    <full-name>Calendar Server developers</full-name>
    <email>dev@bitbucket.calendarserver.org</email>
    <member-uid>__wsanchez__</member-uid>
    <member-uid>__glyph__</member-uid>
    <member-uid>__sagen__</member-uid>
    <member-uid>__cdaboo__</member-uid>
    <member-uid>__dre__</member-uid>
  </record>

  <record type="group">
    <uid>__twisted__</uid>
    <short-name>twisted</short-name>
    <full-name>Twisted Matrix Laboratories</full-name>
    <email>hack@devnull.twistedmatrix.com</email>
    <member-uid>__wsanchez__</member-uid>
    <member-uid>__glyph__</member-uid>
    <member-uid>__exarkun__</member-uid>
    <member-uid>__dreid__</member-uid>
    <member-uid>__dre__</member-uid>
  </record>

  <record type="group">
    <uid>__developers__</uid>
    <short-name>developers</short-name>
    <full-name>All Developers</full-name>
    <member-uid>__calendar-dev__</member-uid>
    <member-uid>__twisted__</member-uid>
    <member-uid>__alyssa__</member-uid>
  </record>

</directory>
"""



##
# Copyright (c) 2013-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Tests for L{twext.python.launchd}.
"""

import json
import os
import plistlib
import socket
import sys

if __name__ == '__main__':
    # This module is loaded as a launchd job by test-cases below; the following
    # code looks up an appropriate function to run.
    testID = sys.argv[1]
    a, b = testID.rsplit(".", 1)
    from twisted.python.reflect import namedAny
    try:
        namedAny(".".join([a, b.replace("test_", "job_")]))()
    finally:
        sys.stdout.flush()
        sys.stderr.flush()
        skt = socket.socket()
        skt.connect(("127.0.0.1", int(os.environ["TESTING_PORT"])))
    sys.exit(0)



try:
    from twext.python.launchd import launchActivateSocket
except ImportError:
    skip = "LaunchD not available."
else:
    skip = False

from twisted.trial.unittest import TestCase
from twisted.python.filepath import FilePath


class CheckInTests(TestCase):
    """
    Integration tests making sure that actual checkin with launchd results in
    the expected values.
    """

    def setUp(self):
        fp = FilePath(self.mktemp())
        fp.makedirs()
        from twisted.internet.protocol import Protocol, Factory
        from twisted.internet import reactor, defer
        d = defer.Deferred()
        class JustLetMeMoveOn(Protocol):
            def connectionMade(self):
                d.callback(None)
                self.transport.abortConnection()
        f = Factory()
        f.protocol = JustLetMeMoveOn
        port = reactor.listenTCP(0, f, interface="127.0.0.1")
        @self.addCleanup
        def goodbyePort():
            return port.stopListening()
        env = dict(os.environ)
        env["TESTING_PORT"] = repr(port.getHost().port)
        self.stdout = fp.child("stdout.txt")
        self.stderr = fp.child("stderr.txt")
        self.launchLabel = ("org.calendarserver.UNIT-TESTS." +
                            str(os.getpid()) + "." + self.id())
        plist = {
            "Label": self.launchLabel,
            "ProgramArguments": [sys.executable, "-m", __name__, self.id()],
            "EnvironmentVariables": env,
            "KeepAlive": False,
            "StandardOutPath": self.stdout.path,
            "StandardErrorPath": self.stderr.path,
            "Sockets": {
                "Awesome": [{"SecureSocketWithKey": "GeneratedSocket"}]
            },
            "RunAtLoad": True,
        }
        self.job = fp.child("job.plist")
        self.job.setContent(plistlib.writePlistToString(plist))
        os.spawnlp(os.P_WAIT, "launchctl", "launchctl", "load", self.job.path)
        return d


    @staticmethod
    def job_test():
        """
        Do something observable in a subprocess.
        """
        sys.stdout.write("Sample Value.")
        sys.stdout.flush()


    def test_test(self):
        """
        Since this test framework is somewhat finicky, let's just make sure
        that a test can complete.
        """
        self.assertEquals("Sample Value.", self.stdout.getContent())


    @staticmethod
    def job_getFDs():
        """
        Check-in via the high-level C{getLaunchDSocketFDs} API, that just gives
        us listening FDs.
        """
        sys.stdout.write(json.dumps(launchActivateSocket("Awesome")))


    def test_getFDs(self):
        """
        L{getLaunchDSocketFDs} returns a Python dictionary mapping the names of
        sockets specified in the property list to lists of integers
        representing FDs.
        """
        sockets = json.loads(self.stdout.getContent())
        self.assertEquals(len(sockets), 1)
        self.assertIsInstance(sockets[0], int)


    def tearDown(self):
        """
        Un-load the launchd job and report any errors it encountered.
        """
        os.spawnlp(os.P_WAIT, "launchctl",
                   "launchctl", "unload", self.job.path)
        err = self.stderr.getContent()
        if 'Traceback' in err:
            self.fail(err)



# -*- test-case-name: twext.who.ldap.test.test_util -*-
##
# Copyright (c) 2013-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from ..idirectory import QueryNotSupportedError, FieldName
from ..expression import (
    CompoundExpression, ExistsExpression, MatchExpression,
    MatchFlags, Operand, BooleanExpression
)
from ._constants import LDAPOperand, LDAPMatchType, LDAPMatchFlags



def ldapQueryStringFromQueryStrings(operand, queryStrings):
    """
    Combines LDAP query strings into a single query string.

    @param operand: An LDAP operand (C{u"&"} or C{u"|"}).
    @type operand: L{unicode}

    @param queryStrings: LDAP query strings.
    @type queryStrings: sequence of L{unicode}
    """
    if len(queryStrings) == 1:
        return queryStrings[0]

    elif len(queryStrings) > 1:
        queryTokens = []
        queryTokens.append(u"(")
        queryTokens.append(operand)
        queryTokens.extend(queryStrings)
        queryTokens.append(u")")
        return u"".join(queryTokens)

    else:
        return u""



def ldapQueryStringFromMatchExpression(
    expression, fieldNameToAttributesMap, recordTypeSchemas
):
    """
    Generates an LDAP query string from a match expression.

    @param expression: A match expression.
    @type expression: L{MatchExpression}

    @param fieldNameToAttributesMap: A mapping from field names to native LDAP
        attribute names.
    @type fieldNameToAttributesMap: L{dict} with L{FieldName} keys and sequence
        of L{unicode} values.

    @param recordTypeSchemas: Schema information for record types.
    @type recordTypeSchemas: mapping from L{NamedConstant} to
        L{RecordTypeSchema}

    @return: An LDAP query string.
    @rtype: L{unicode}

    @raises QueryNotSupportedError: If the expression's match type is unknown,
        or if the expresion references an unknown field name (meaning a field
        name not in C{fieldNameToAttributeMap}).
    """

    matchType = LDAPMatchType.fromMatchType(expression.matchType)
    if matchType is None:
        raise QueryNotSupportedError(
            "Unknown match type: {0}".format(matchType)
        )

    flags = tuple(expression.flags)

    if MatchFlags.NOT in flags:
        notOp = LDAPMatchFlags.NOT.value
        operand = LDAPOperand.AND.value
    else:
        notOp = LDAPMatchFlags.none.value
        operand = LDAPOperand.OR.value

    # FIXME: It doesn't look like LDAP queries can be case sensitive.
    # This would mean that it's up to the callers to filter out the false
    # positives...
    #
    # if MatchFlags.caseInsensitive not in flags:
    #     raise NotImplementedError("Need to handle case sensitive")

    fieldName = expression.fieldName

    if fieldName is FieldName.recordType:
        if matchType is not LDAPMatchType.equals:
            raise QueryNotSupportedError(
                "Match type for record type must be equals, not {0!r}"
                .format(expression.matchType)
            )

        try:
            schema = recordTypeSchemas[expression.fieldValue]
        except KeyError:
            raise QueryNotSupportedError(
                "Unmapped record type: {0}".format(expression.fieldValue)
            )

        if notOp:
            valueOperand = LDAPOperand.OR.value
        else:
            valueOperand = LDAPOperand.AND.value

        queryStrings = [
            matchType.queryString.format(
                notOp=notOp, attribute=attribute, value=value,
            )
            for attribute, value in schema.attributes
        ]

        return ldapQueryStringFromQueryStrings(valueOperand, queryStrings)

    else:
        try:
            attributes = fieldNameToAttributesMap[fieldName]
        except KeyError:
            raise QueryNotSupportedError(
                "Unmapped field name: {0}".format(expression.fieldName)
            )

        # Covert to unicode and escape special LDAP query characters
        value = unicode(expression.fieldValue).translate(LDAP_QUOTING_TABLE)

        # Compose an query using each of the LDAP attributes cooresponding to
        # the target field name.

        queryStrings = [
            matchType.queryString.format(
                notOp=notOp, attribute=attribute, value=value
            )
            for attribute in attributes
        ]

        return ldapQueryStringFromQueryStrings(operand, queryStrings)

    raise AssertionError("We shouldn't be here.")



def ldapQueryStringFromExistsExpression(
    expression, fieldNameToAttributesMap, recordTypeSchemas
):
    """
    Generates an LDAP query string from an exists expression.

    @param expression: An exists expression.
    @type expression: L{ExistsExpression}

    @param fieldNameToAttributesMap: A mapping from field names to native LDAP
        attribute names.
    @type fieldNameToAttributesMap: L{dict} with L{FieldName} keys and sequence
        of L{unicode} values.

    @param recordTypeSchemas: Schema information for record types.
    @type recordTypeSchemas: mapping from L{NamedConstant} to
        L{RecordTypeSchema}

    @return: An LDAP query string.
    @rtype: L{unicode}

    @raises QueryNotSupportedError: if the expresion references an unknown
        field name (meaning a field name not in C{fieldNameToAttributeMap}).
    """

    fieldName = expression.fieldName

    try:
        attributes = fieldNameToAttributesMap[fieldName]
    except KeyError:
        raise QueryNotSupportedError(
            "Unmapped field name: {0}".format(expression.fieldName)
        )

    queryStrings = [
        u"({attribute}=*)".format(attribute=attribute)
        for attribute in attributes
    ]

    operand = LDAPOperand.OR.value
    return ldapQueryStringFromQueryStrings(operand, queryStrings)



def ldapQueryStringFromBooleanExpression(
    expression, fieldNameToAttributesMap, recordTypeSchemas
):
    """
    Generates an LDAP query string from a boolean expression.

    @param expression: An boolean expression.
    @type expression: L{BooleanExpression}

    @param fieldNameToAttributesMap: A mapping from field names to native LDAP
        attribute names.
    @type fieldNameToAttributesMap: L{dict} with L{FieldName} keys and sequence
        of L{unicode} values.

    @param recordTypeSchemas: Schema information for record types.
    @type recordTypeSchemas: mapping from L{NamedConstant} to
        L{RecordTypeSchema}

    @return: An LDAP query string.
    @rtype: L{unicode}

    @raises QueryNotSupportedError: if the expresion references an unknown
        field name (meaning a field name not in C{fieldNameToAttributeMap}).
    """

    fieldName = expression.fieldName

    try:
        attributes = fieldNameToAttributesMap[fieldName]
    except KeyError:
        raise QueryNotSupportedError(
            "Unmapped field name: {0}".format(expression.fieldName)
        )

    queryStrings = []
    for attribute in attributes:
        if ":" in attribute:
            attribute, trueValue = attribute.split(":")
        else:
            trueValue = "true"

        queryStrings.append(
            u"({attribute}={trueValue})".format(
                attribute=attribute,
                trueValue=trueValue
            )
        )


    operand = LDAPOperand.OR.value
    return ldapQueryStringFromQueryStrings(operand, queryStrings)



def ldapQueryStringFromCompoundExpression(
    expression, fieldNameToAttributesMap, recordTypeSchemas
):
    """
    Generates an LDAP query string from a compound expression.

    @param expression: A compound expression.
    @type expression: L{MatchExpression}

    @param fieldNameToAttributesMap: A mapping from field names to native LDAP
        attribute names.
    @type fieldNameToAttributesMap: L{dict} with L{FieldName} keys and sequence
        of L{unicode} values.

    @param recordTypeSchemas: Schema information for record types.
    @type recordTypeSchemas: mapping from L{NamedConstant} to
        L{RecordTypeSchema}

    @return: An LDAP query string.
    @rtype: L{unicode}

    @raises QueryNotSupportedError: If any sub-expression cannot be converted
        to an LDAP query.
    """
    if expression.operand is Operand.AND:
        operand = LDAPOperand.AND.value

    elif expression.operand is Operand.OR:
        operand = LDAPOperand.OR.value

    queryStrings = [
        ldapQueryStringFromExpression(
            subExpression,
            fieldNameToAttributesMap, recordTypeSchemas
        )
        for subExpression in expression.expressions
    ]

    return ldapQueryStringFromQueryStrings(operand, queryStrings)



def ldapQueryStringFromExpression(
    expression, fieldNameToAttributesMap, recordTypeSchemas
):
    """
    Converts an expression into an LDAP query string.

    @param expression: An expression.
    @type expression: L{MatchExpression} or L{CompoundExpression}

    @param fieldNameToAttributesMap: A mapping from field names to native LDAP
        attribute names.
    @type fieldNameToAttributesMap: L{dict} with L{FieldName} keys and sequence
        of L{unicode} values.

    @param recordTypeSchemas: Schema information for record types.
    @type recordTypeSchemas: mapping from L{NamedConstant} to
        L{RecordTypeSchema}

    @return: An LDAP query string.
    @rtype: L{unicode}

    @raises QueryNotSupportedError: If the expression cannot be converted to an
        LDAP query.
    """

    if isinstance(expression, MatchExpression):
        return ldapQueryStringFromMatchExpression(
            expression, fieldNameToAttributesMap, recordTypeSchemas
        )

    if isinstance(expression, BooleanExpression):
        return ldapQueryStringFromBooleanExpression(
            expression, fieldNameToAttributesMap, recordTypeSchemas
        )

    if isinstance(expression, ExistsExpression):
        return ldapQueryStringFromExistsExpression(
            expression, fieldNameToAttributesMap, recordTypeSchemas
        )

    if isinstance(expression, CompoundExpression):
        return ldapQueryStringFromCompoundExpression(
            expression, fieldNameToAttributesMap, recordTypeSchemas
        )

    raise QueryNotSupportedError(
        "Unknown expression type: {0!r}".format(expression)
    )


LDAP_QUOTING_TABLE = {
    ord(u"\\"): u"\\5C",
    ord(u"/"): u"\\2F",

    ord(u"("): u"\\28",
    ord(u")"): u"\\29",
    ord(u"*"): u"\\2A",

    ord(u"<"): u"\\3C",
    ord(u"="): u"\\3D",
    ord(u">"): u"\\3E",
    ord(u"~"): u"\\7E",

    ord(u"&"): u"\\26",
    ord(u"|"): u"\\7C",

    ord(u"\0"): u"\\00",
}



# -*- test-case-name: twext.enterprise.test.test_fixtures -*-
##
# Copyright (c) 2012-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Fixtures for testing code that uses ADBAPI2.
"""

import sqlite3
from Queue import Empty
from itertools import count

from zope.interface import implementer
from zope.interface.verify import verifyClass

from twisted.internet.interfaces import IReactorThreads
from twisted.python.threadpool import ThreadPool

from twisted.internet.task import Clock

from twext.enterprise.adbapi2 import ConnectionPool
from twext.enterprise.ienterprise import SQLITE_DIALECT
from twext.enterprise.ienterprise import POSTGRES_DIALECT
from twext.enterprise.adbapi2 import DEFAULT_PARAM_STYLE
from twext.internet.threadutils import ThreadHolder



def buildConnectionPool(testCase, schemaText="", dialect=SQLITE_DIALECT):
    """
    Build a L{ConnectionPool} for testing purposes, with the given C{testCase}.

    @param testCase: the test case to attach the resulting L{ConnectionPool}
        to.
    @type testCase: L{twisted.trial.unittest.TestCase}

    @param schemaText: The text of the schema with which to initialize the
        database.
    @type schemaText: L{str}

    @return: a L{ConnectionPool} service whose C{startService} method has
        already been invoked.
    @rtype: L{ConnectionPool}
    """
    sqlitename = testCase.mktemp()
    seqs = {}

    def connectionFactory(label=testCase.id()):
        conn = sqlite3.connect(sqlitename, isolation_level=None)

        def nextval(seq):
            result = seqs[seq] = seqs.get(seq, 0) + 1
            return result

        conn.create_function("nextval", 1, nextval)
        return conn

    con = connectionFactory()
    con.executescript(schemaText)
    con.commit()
    pool = ConnectionPool(connectionFactory, paramstyle="numeric",
                          dialect=SQLITE_DIALECT)
    pool.startService()
    testCase.addCleanup(pool.stopService)
    return pool



def resultOf(deferred, propagate=False):
    """
    Add a callback and errback which will capture the result of a L{Deferred}
    in a list, and return that list.  If C{propagate} is True, pass through the
    results.
    """
    results = []

    if propagate:
        def cb(r):
            results.append(r)
            return r
    else:
        cb = results.append

    deferred.addBoth(cb)
    return results



class FakeThreadHolder(ThreadHolder):
    """
    Run things to submitted this ThreadHolder on the main thread, so that
    execution is easier to control.
    """

    def __init__(self, test):
        super(FakeThreadHolder, self).__init__(self)
        self.test = test
        self.started = False
        self.stopped = False
        self._workerIsRunning = False


    def start(self):
        self.started = True
        return super(FakeThreadHolder, self).start()


    def stop(self):
        result = super(FakeThreadHolder, self).stop()
        self.stopped = True
        return result


    @property
    def _get_q(self):
        return self._q_


    @_get_q.setter
    def _q(self, newq):
        if newq is not None:
            oget = newq.get
            newq.get = lambda: oget(timeout=0)
            oput = newq.put

            def putit(x):
                p = oput(x)
                if not self.test.paused:
                    self.flush()
                return p

            newq.put = putit

        self._q_ = newq


    def callFromThread(self, f, *a, **k):
        result = f(*a, **k)
        return result


    def callInThread(self, f, *a, **k):
        """
        This should be called only once, to start the worker function that
        dedicates a thread to this L{ThreadHolder}.
        """
        self._workerIsRunning = True


    def flush(self):
        """
        Fire all deferreds previously returned from submit.
        """
        try:
            while self._workerIsRunning and self._qpull():
                pass
            else:
                self._workerIsRunning = False
        except Empty:
            pass



@implementer(IReactorThreads)
class ClockWithThreads(Clock):
    """
    A testing reactor that supplies L{IReactorTime} and L{IReactorThreads}.
    """

    def __init__(self):
        super(ClockWithThreads, self).__init__()
        self._pool = ThreadPool()


    def getThreadPool(self):
        """
        Get the threadpool.
        """
        return self._pool


    def suggestThreadPoolSize(self, size):
        """
        Approximate the behavior of a "real" reactor.
        """
        self._pool.adjustPoolsize(maxthreads=size)


    def callInThread(self, thunk, *a, **kw):
        """
        No implementation.
        """

    def callFromThread(self, thunk, *a, **kw):
        """
        No implementation.
        """

verifyClass(IReactorThreads, ClockWithThreads)



class ConnectionPoolHelper(object):
    """
    Connection pool setting-up facilities for tests that need a
    L{ConnectionPool}.
    """

    dialect = POSTGRES_DIALECT
    paramstyle = DEFAULT_PARAM_STYLE

    def setUp(self, test=None, connect=None):
        """
        Support inheritance by L{TestCase} classes.
        """
        if test is None:
            test = self
        if connect is None:
            self.factory = ConnectionFactory()
            connect = self.factory.connect
        self.connect = connect
        self.paused = False
        self.holders = []
        self.pool = ConnectionPool(
            connect,
            maxConnections=2,
            dialect=self.dialect,
            paramstyle=self.paramstyle
        )
        self.pool._createHolder = self.makeAHolder
        self.clock = self.pool.reactor = ClockWithThreads()
        self.pool.startService()
        test.addCleanup(self.flushHolders)


    def flushHolders(self):
        """
        Flush all pending C{submit}s since C{pauseHolders} was called.  This
        makes sure the service is stopped and the fake ThreadHolders are all
        executing their queues so failed tests can exit cleanly.
        """
        self.paused = False
        for holder in self.holders:
            holder.flush()


    def pauseHolders(self):
        """
        Pause all L{FakeThreadHolder}s, causing C{submit} to return an unfired
        L{Deferred}.
        """
        self.paused = True


    def makeAHolder(self):
        """
        Make a ThreadHolder-alike.
        """
        fth = FakeThreadHolder(self)
        self.holders.append(fth)
        return fth


    def resultOf(self, it):
        return resultOf(it)


    def createTransaction(self):
        return self.pool.connection()


    def translateError(self, err):
        return err



class SteppablePoolHelper(ConnectionPoolHelper):
    """
    A version of L{ConnectionPoolHelper} that can set up a connection pool
    capable of firing all its L{Deferred}s on demand, synchronously, by using
    SQLite.
    """
    dialect = SQLITE_DIALECT
    paramstyle = sqlite3.paramstyle

    def __init__(self, schema):
        self.schema = schema


    def setUp(self, test):
        connect = synchronousConnectionFactory(test)
        con = connect()
        cur = con.cursor()
        cur.executescript(self.schema)
        con.commit()
        super(SteppablePoolHelper, self).setUp(test, connect)


    def rows(self, sql):
        """
        Get some rows from the database to compare in a test.
        """
        con = self.connect()
        cur = con.cursor()
        cur.execute(sql)
        result = cur.fetchall()
        con.commit()
        return result



def synchronousConnectionFactory(test):
    tmpdb = test.mktemp()

    def connect():
        return sqlite3.connect(tmpdb, isolation_level=None)

    return connect



class Child(object):
    """
    An object with a L{Parent}, in its list of C{children}.
    """
    def __init__(self, parent):
        self.closed = False
        self.parent = parent
        self.parent.children.append(self)


    def close(self):
        if self.parent._closeFailQueue:
            raise self.parent._closeFailQueue.pop(0)
        self.closed = True



class Parent(object):
    """
    An object with a list of L{Child}ren.
    """

    def __init__(self):
        self.children = []
        self._closeFailQueue = []


    def childCloseWillFail(self, exception):
        """
        Closing children of this object will result in the given exception.

        @see: L{ConnectionFactory}
        """
        self._closeFailQueue.append(exception)



class FakeConnection(Parent, Child):
    """
    Fake Stand-in for DB-API 2.0 connection.

    @ivar executions: the number of statements which have been executed.

    """

    executions = 0

    def __init__(self, factory):
        """
        Initialize list of cursors
        """
        Parent.__init__(self)
        Child.__init__(self, factory)
        self.id = factory.idcounter.next()
        self._executeFailQueue = []
        self._commitCount = 0
        self._rollbackCount = 0


    def executeWillFail(self, thunk):
        """
        The next call to L{FakeCursor.execute} will fail with an exception
        returned from the given callable.
        """
        self._executeFailQueue.append(thunk)


    @property
    def cursors(self):
        "Alias to make tests more readable."
        return self.children


    def cursor(self):
        return FakeCursor(self)


    def commit(self):
        self._commitCount += 1
        if self.parent.commitFail:
            self.parent.commitFail = False
            raise CommitFail()


    def rollback(self):
        self._rollbackCount += 1
        if self.parent.rollbackFail:
            self.parent.rollbackFail = False
            raise RollbackFail()



class RollbackFail(Exception):
    """
    Sample rollback-failure exception.
    """



class CommitFail(Exception):
    """
    Sample Commit-failure exception.
    """



class FakeCursor(Child):
    """
    Fake stand-in for a DB-API 2.0 cursor.
    """
    def __init__(self, connection):
        Child.__init__(self, connection)
        self.rowcount = 0
        # not entirely correct, but all we care about is its truth value.
        self.description = False
        self.variables = []
        self.allExecutions = []


    @property
    def connection(self):
        "Alias to make tests more readable."
        return self.parent


    def execute(self, sql, args=()):
        if self.connection.closed:
            raise FakeConnectionError
        self.connection.executions += 1
        if self.connection._executeFailQueue:
            raise self.connection._executeFailQueue.pop(0)()
        self.allExecutions.append((sql, args))
        self.sql = sql
        factory = self.connection.parent
        self.description = factory.hasResults
        if factory.hasResults and factory.shouldUpdateRowcount:
            self.rowcount = 1
        else:
            self.rowcount = 0
        return


    def var(self, type, *args):
        """
        Return a database variable in the style of the cx_Oracle bindings.
        """
        v = FakeVariable(self, type, args)
        self.variables.append(v)
        return v


    def fetchall(self):
        """
        Just echo the SQL that was executed in the last query.
        """
        if self.connection.parent.hasResults:
            return [[self.connection.id, self.sql]]
        if self.description:
            return []
        return None



class FakeVariable(object):
    def __init__(self, cursor, type, args):
        self.cursor = cursor
        self.type = type
        self.args = args


    def getvalue(self):
        vv = self.cursor.connection.parent.varvals
        if vv:
            return vv.pop(0)
        return self.cursor.variables.index(self) + 300


    def __reduce__(self):
        raise RuntimeError("Not pickleable (since oracle vars aren't)")



class ConnectionFactory(Parent):
    """
    A factory for L{FakeConnection} objects.

    @ivar shouldUpdateRowcount: Should C{execute} on cursors produced by
        connections produced by this factory update their C{rowcount} or just
        their C{description} attribute?

    @ivar hasResults: should cursors produced by connections by this factory
        have any results returned by C{fetchall()}?
    """

    rollbackFail = False
    commitFail = False

    def __init__(self, shouldUpdateRowcount=True, hasResults=True):
        Parent.__init__(self)
        self.idcounter = count(1)
        self._connectResultQueue = []
        self.defaultConnect()
        self.varvals = []
        self.shouldUpdateRowcount = shouldUpdateRowcount
        self.hasResults = hasResults


    @property
    def connections(self):
        "Alias to make tests more readable."
        return self.children


    def connect(self):
        """
        Implement the C{ConnectionFactory} callable expected by
        L{ConnectionPool}.
        """
        if self._connectResultQueue:
            thunk = self._connectResultQueue.pop(0)
        else:
            thunk = self._default
        return thunk()


    def willConnect(self):
        """
        Used by tests to queue a successful result for connect().
        """
        def thunk():
            return FakeConnection(self)
        self._connectResultQueue.append(thunk)


    def willConnectTo(self):
        """
        Queue a successful result for connect() and immediately add it as a
        child to this L{ConnectionFactory}.

        @return: a connection object
        @rtype: L{FakeConnection}
        """
        aConnection = FakeConnection(self)

        def thunk():
            return aConnection

        self._connectResultQueue.append(thunk)
        return aConnection


    def willFail(self):
        """
        Used by tests to queue a successful result for connect().
        """
        def thunk():
            raise FakeConnectionError()
        self._connectResultQueue.append(thunk)


    def defaultConnect(self):
        """
        By default, connection attempts will succeed.
        """
        self.willConnect()
        self._default = self._connectResultQueue.pop()


    def defaultFail(self):
        """
        By default, connection attempts will fail.
        """
        self.willFail()
        self._default = self._connectResultQueue.pop()



class FakeConnectionError(Exception):
    """
    Synthetic error that might occur during connection.
    """



# -*- test-case-name: twext.enterprise.test.test_locking -*-
##
# Copyright (c) 2012-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Utilities to restrict concurrency based on mutual exclusion.
"""

from twext.enterprise.dal.model import Table
from twext.enterprise.dal.model import SQLType
from twext.enterprise.dal.model import Constraint
from twext.enterprise.dal.syntax import SchemaSyntax
from twext.enterprise.dal.model import Schema
from twext.enterprise.dal.record import Record
from twext.enterprise.dal.record import fromTable


class AlreadyUnlocked(Exception):
    """
    The lock you were trying to unlock was already unlocked.
    """



class LockTimeout(Exception):
    """
    The lock you were trying to lock was already locked causing a timeout.
    """



def makeLockSchema(inSchema):
    """
    Create a self-contained schema just for L{Locker} use, in C{inSchema}.

    @param inSchema: a L{Schema} to add the locks table to.
    @type inSchema: L{Schema}

    @return: inSchema
    """
    LockTable = Table(inSchema, "NAMED_LOCK")

    LockTable.addColumn("LOCK_NAME", SQLType("varchar", 255))
    LockTable.tableConstraint(Constraint.NOT_NULL, ["LOCK_NAME"])
    LockTable.tableConstraint(Constraint.UNIQUE, ["LOCK_NAME"])
    LockTable.primaryKey = [LockTable.columnNamed("LOCK_NAME")]

    return inSchema

LockSchema = SchemaSyntax(makeLockSchema(Schema(__file__)))



class NamedLock(Record, fromTable(LockSchema.NAMED_LOCK)):
    """
    An L{AcquiredLock} lock against a shared data store that the current
    process holds via the referenced transaction.
    """

    @classmethod
    def acquire(cls, txn, name):
        """
        Acquire a lock with the given name.

        @param name: The name of the lock to acquire.  Against the same store,
            no two locks may be acquired.
        @type name: L{unicode}

        @return: a L{Deferred} that fires with an L{AcquiredLock} when the lock
            has fired, or fails when the lock has not been acquired.
        """
        def autoRelease(self):
            txn.preCommit(lambda: self.release(True))
            return self

        def lockFailed(f):
            raise LockTimeout(name)

        d = cls.create(txn, lockName=name)
        d.addCallback(autoRelease)
        d.addErrback(lockFailed)
        return d


    def release(self, ignoreAlreadyUnlocked=False):
        """
        Release this lock.

        @param ignoreAlreadyUnlocked: If you don't care about the current
            status of this lock, and just want to release it if it is still
            acquired, pass this parameter as L{True}.  Otherwise this method
            will raise an exception if it is invoked when the lock has already
            been released.

        @raise: L{AlreadyUnlocked}

        @return: A L{Deferred} that fires with L{None} when the lock has been
            unlocked.
        """
        return self.delete()



##
# Copyright (c) 2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Socket file implementation for MaxAccept
"""

__all__ = [
    "MaxAcceptSocketFileServer",
]


from twisted.application import service
from twisted.internet import endpoints
from twisted.internet.defer import inlineCallbacks

from twext.python.log import Logger

log = Logger()



def maxAcceptDoRead(self):
    self.numberAccepts = min(
        self.factory.maxRequests - self.factory.outstandingRequests,
        self.factory.maxAccepts
    )
    self.realDoRead()


class MaxAcceptSocketFileServer(service.Service):
    """
    Socket File server

    @ivar myPort: When running, this is set to the L{IListeningPort} being
        managed by this service.
    """

    def __init__(self, protocolFactory, address, backlog=None):
        self.protocolFactory = protocolFactory
        self.protocolFactory.myServer = self
        self.address = address
        self.backlog = backlog
        self.myPort = None


    @inlineCallbacks
    def startService(self):
        from twisted.internet import reactor
        endpoint = endpoints.UNIXServerEndpoint(
            reactor, self.address, backlog=self.backlog, wantPID=1
        )
        self.myPort = yield endpoint.listen(self.protocolFactory)

        # intercept doRead() to set numberAccepts
        self.myPort.realDoRead = self.myPort.doRead
        self.myPort.doRead = maxAcceptDoRead.__get__(
            self.myPort, self.myPort.__class__
        )


    @inlineCallbacks
    def stopService(self):
        """
        Wait for outstanding requests to finish

        @return: a Deferred which fires when all outstanding requests are
            complete
        """
        if self.myPort is not None:
            yield self.myPort.stopListening()

        if hasattr(self.protocolFactory, "allConnectionsClosed"):
            yield self.protocolFactory.allConnectionsClosed()



##
# Copyright (c) 2013-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Directory service expression tests.
"""

from twisted.trial import unittest
from uuid import UUID

from ..idirectory import FieldName
from ..expression import MatchExpression, MatchType, MatchFlags



class MatchFlagsTest(unittest.TestCase):
    """
    Tests for L{MatchFlags}.
    """

    def test_predicator_none(self):
        """
        Predicator for flags without L{MatchFlags.NOT} does not invert.
        """
        predicator = MatchFlags.predicator(MatchFlags.none)

        for boolean in (True, False):
            self.assertEquals(bool(predicator(boolean)), boolean)


    def test_predicator_NOT(self):
        """
        Predicator for flags with L{MatchFlags.NOT} does not invert.
        """
        predicator = MatchFlags.predicator(MatchFlags.NOT)

        for boolean in (True, False):
            self.assertNotEquals(bool(predicator(boolean)), boolean)


    def test_normalizer_none(self):
        """
        Normalizer for flags without L{MatchFlags.caseInsensitive} does not
        lowercase.
        """
        normalizer = MatchFlags.normalizer(MatchFlags.none)

        self.assertEquals(normalizer(u"ThInGo"), u"ThInGo")


    def test_normalizer_insensitive(self):
        """
        Normalizer for flags with L{MatchFlags.caseInsensitive} lowercases.
        """
        normalizer = MatchFlags.normalizer(MatchFlags.caseInsensitive)

        self.assertEquals(normalizer(u"ThInGo"), u"thingo")



class MatchExpressionTest(unittest.TestCase):
    """
    Tests for L{MatchExpression}.
    """

    def test_initBadType_name(self):
        """
        L{MatchExpression.__init__} raises if the field name is not a
        L{NamedConstant}.
        """
        self.assertRaises(
            TypeError,
            MatchExpression, u"uid", u"some value"
        )


    def test_initBadType_value(self):
        """
        L{MatchExpression.__init__} raises if the field value doesn't match the
        expected type for the field.
        """
        # guid field expects a UUID, not a string.
        self.assertRaises(
            TypeError,
            MatchExpression,
            FieldName.guid, u"00000000-0000-0000-0000-000000000000"
        )


    def test_repr_name(self):
        """
        L{MatchExpression.__repr__} emits field name and value.
        """
        self.assertEquals(
            "<MatchExpression: u'full names' equals u'Wilfredo Sanchez'>",
            repr(MatchExpression(
                FieldName.fullNames,
                u"Wilfredo Sanchez",
            )),
        )


    def test_repr_type(self):
        """
        L{MatchExpression.__repr__} emits match type.
        """
        self.assertEquals(
            "<MatchExpression: u'full names' contains u'Sanchez'>",
            repr(MatchExpression(
                FieldName.fullNames,
                u"Sanchez",
                matchType=MatchType.contains,
            )),
        )


    def test_repr_flags(self):
        """
        L{MatchExpression.__repr__} emits flags.
        """
        self.assertEquals(
            "<MatchExpression: u'full names' starts with u'Wilfredo' (not)>",
            repr(MatchExpression(
                FieldName.fullNames,
                u"Wilfredo",
                matchType=MatchType.startsWith,
                flags=MatchFlags.NOT,
            )),
        )
        self.assertEquals(
            "<MatchExpression: u'full names' starts with u'Wilfredo' "
            "(case insensitive|not)>",
            repr(MatchExpression(
                FieldName.fullNames,
                u"Wilfredo",
                matchType=MatchType.startsWith,
                flags=(MatchFlags.NOT | MatchFlags.caseInsensitive),
            )),
        )


    def test_match_equals(self):
        """
        L{MatchExpression.match} with L{MatchType.equals}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"Morgen",
            matchType=MatchType.equals,
        )
        self.assertTrue(expression.match(u"Morgen"))
        self.assertFalse(expression.match(u"Wilfredo"))


    def test_match_equals_not(self):
        """
        L{MatchExpression.match} with L{MatchType.equals} and
        L{MatchFlags.NOT}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"Morgen",
            matchType=MatchType.equals,
            flags=MatchFlags.NOT,
        )
        self.assertFalse(expression.match(u"Morgen"))
        self.assertTrue(expression.match(u"Wilfredo"))


    def test_match_startsWith(self):
        """
        L{MatchExpression.match} with L{MatchType.startsWith}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"Mor",
            matchType=MatchType.startsWith,
        )
        self.assertTrue(expression.match(u"Morgen"))
        self.assertFalse(expression.match(u"Wilfredo"))


    def test_match_startsWith_not(self):
        """
        L{MatchExpression.match} with L{MatchType.startsWith} and
        L{MatchFlags.NOT}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"Mor",
            matchType=MatchType.startsWith,
            flags=MatchFlags.NOT,
        )
        self.assertFalse(expression.match(u"Morgen"))
        self.assertTrue(expression.match(u"Wilfredo"))


    def test_match_endsWith(self):
        """
        L{MatchExpression.match} with L{MatchType.endsWith}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"gen",
            matchType=MatchType.endsWith,
        )
        self.assertTrue(expression.match(u"Morgen"))
        self.assertFalse(expression.match(u"Wilfredo"))


    def test_match_endsWith_not(self):
        """
        L{MatchExpression.match} with L{MatchType.endsWith} and
        L{MatchFlags.NOT}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"gen",
            matchType=MatchType.endsWith,
            flags=MatchFlags.NOT,
        )
        self.assertFalse(expression.match(u"Morgen"))
        self.assertTrue(expression.match(u"Wilfredo"))


    def test_match_contains(self):
        """
        L{MatchExpression.match} with L{MatchType.contains}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"org",
            matchType=MatchType.contains,
        )
        self.assertTrue(expression.match(u"Morgen"))
        self.assertFalse(expression.match(u"Wilfredo"))


    def test_match_contains_not(self):
        """
        L{MatchExpression.match} with L{MatchType.contains} and
        L{MatchFlags.NOT}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"org",
            matchType=MatchType.contains,
            flags=MatchFlags.NOT,
        )
        self.assertFalse(expression.match(u"Morgen"))
        self.assertTrue(expression.match(u"Wilfredo"))


    def test_match_lessThan(self):
        """
        L{MatchExpression.match} with L{MatchType.lessThan}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"Morgen",
            matchType=MatchType.lessThan,
        )
        self.assertTrue(expression.match(u"Glyph"))  # Sorry, Glyph
        self.assertFalse(expression.match(u"Wilfredo"))


    def test_match_lessThan_not(self):
        """
        L{MatchExpression.match} with L{MatchType.lessThan} and
        L{MatchFlags.NOT}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"Morgen",
            matchType=MatchType.lessThan,
            flags=MatchFlags.NOT,
        )
        self.assertFalse(expression.match(u"Glyph"))
        self.assertTrue(expression.match(u"Wilfredo"))


    def test_match_greaterThan(self):
        """
        L{MatchExpression.match} with L{MatchType.greaterThan}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"Morgen",
            matchType=MatchType.greaterThan,
        )
        self.assertTrue(expression.match(u"Wilfredo"))  # Woot!
        self.assertFalse(expression.match(u"Glyph"))


    def test_match_greaterThan_not(self):
        """
        L{MatchExpression.match} with L{MatchType.greaterThan} and
        L{MatchFlags.NOT}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"Morgen",
            matchType=MatchType.greaterThan,
            flags=MatchFlags.NOT,
        )
        self.assertFalse(expression.match(u"Wilfredo"))
        self.assertTrue(expression.match(u"Glyph"))


    def test_match_lessThanOrEqualTo(self):
        """
        L{MatchExpression.match} with L{MatchType.lessThanOrEqualTo}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"Morgen",
            matchType=MatchType.lessThanOrEqualTo,
        )
        self.assertTrue(expression.match(u"Glyph"))
        self.assertTrue(expression.match(u"Morgen"))
        self.assertFalse(expression.match(u"Wilfredo"))


    def test_match_lessThanOrEqualTo_not(self):
        """
        L{MatchExpression.match} with L{MatchType.lessThanOrEqualTo} and
        L{MatchFlags.NOT}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"Morgen",
            matchType=MatchType.lessThanOrEqualTo,
            flags=MatchFlags.NOT,
        )
        self.assertFalse(expression.match(u"Glyph"))
        self.assertFalse(expression.match(u"Morgen"))
        self.assertTrue(expression.match(u"Wilfredo"))


    def test_match_greaterThanOrEqualTo(self):
        """
        L{MatchExpression.match} with L{MatchType.greaterThanOrEqualTo}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"Morgen",
            matchType=MatchType.greaterThanOrEqualTo,
        )
        self.assertTrue(expression.match(u"Wilfredo"))
        self.assertTrue(expression.match(u"Morgen"))
        self.assertFalse(expression.match(u"Glyph"))


    def test_match_greaterThanOrEqualTo_not(self):
        """
        L{MatchExpression.match} with L{MatchType.greaterThanOrEqualTo} and
        L{MatchFlags.NOT}.
        """
        expression = MatchExpression(
            FieldName.fullNames, u"Morgen",
            matchType=MatchType.greaterThanOrEqualTo,
            flags=MatchFlags.NOT,
        )
        self.assertFalse(expression.match(u"Wilfredo"))
        self.assertFalse(expression.match(u"Morgen"))
        self.assertTrue(expression.match(u"Glyph"))


    def test_non_unicode_value(self):
        """
        L{MatchExpression.match} with L{MatchType.equals},
        L{MatchFlags.caseInsensitive} and a UUID value (which cannot be
        lower()'d)
        """
        uuidValue = UUID("95F868E5-EFBD-4BFE-8DFB-25C3BC5CCDDA")
        expression = MatchExpression(
            FieldName.guid, uuidValue,
            matchType=MatchType.equals,
            flags=MatchFlags.caseInsensitive,

        )
        self.assertFalse(expression.match(u"Bogus"))
        self.assertTrue(expression.match(uuidValue))



##
# Copyright (c) 2012-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Tests for L{twext.internet.adaptendpoint}.
"""

from zope.interface.verify import verifyObject

from twext.internet.adaptendpoint import connect
from twisted.internet.defer import Deferred, CancelledError
from twisted.python.failure import Failure

from twisted.internet.protocol import ClientFactory, Protocol
from twisted.internet.interfaces import IConnector
from twisted.trial.unittest import TestCase

class names(object):
    def __init__(self, **kw):
        self.__dict__.update(kw)



class RecordingProtocol(Protocol, object):
    def __init__(self):
        super(RecordingProtocol, self).__init__()
        self.made = []
        self.data = []
        self.lost = []


    def connectionMade(self):
        self.made.append(self.transport)


    def dataReceived(self, data):
        self.data.append(data)


    def connectionLost(self, why):
        self.lost.append(why)



class RecordingClientFactory(ClientFactory):
    """
    L{ClientFactory} subclass that records the things that happen to it.
    """

    def __init__(self):
        """
        Create some records of things that are about to happen.
        """
        self.starts = []
        self.built = []
        self.fails = []
        self.lost = []


    def startedConnecting(self, ctr):
        self.starts.append(ctr)


    def clientConnectionFailed(self, ctr, reason):
        self.fails.append(names(connector=ctr, reason=reason))


    def clientConnectionLost(self, ctr, reason):
        self.lost.append(names(connector=ctr, reason=reason))


    def buildProtocol(self, addr):
        b = RecordingProtocol()
        self.built.append(names(protocol=b, addr=addr))
        return b



class RecordingEndpoint(object):

    def __init__(self):
        self.attempts = []


    def connect(self, factory):
        d = Deferred()
        self.attempts.append(names(deferred=d, factory=factory))
        return d



class RecordingTransport(object):

    def __init__(self):
        self.lose = []


    def loseConnection(self):
        self.lose.append(self)



class AdaptEndpointTests(TestCase):
    """
    Tests for L{connect} and the objects that it coordinates.
    """

    def setUp(self):
        self.factory = RecordingClientFactory()
        self.endpoint = RecordingEndpoint()
        self.connector = connect(self.endpoint, self.factory)


    def connectionSucceeds(self, addr=object()):
        """
        The most recent connection attempt succeeds, returning the L{ITransport}
        provider produced by its success.
        """
        transport = RecordingTransport()
        attempt = self.endpoint.attempts[-1]
        proto = attempt.factory.buildProtocol(addr)
        proto.makeConnection(transport)
        transport.protocol = proto
        attempt.deferred.callback(proto)
        return transport


    def connectionFails(self, reason):
        """
        The most recent in-progress connection fails.
        """
        self.endpoint.attempts[-1].deferred.errback(reason)


    def test_connectStartsConnection(self):
        """
        When used with a successful endpoint, L{connect} will simulate all
        aspects of the connection process; C{buildProtocol}, C{connectionMade},
        C{dataReceived}.
        """
        self.assertIdentical(self.connector.getDestination(), self.endpoint)
        verifyObject(IConnector, self.connector)
        self.assertEqual(self.factory.starts, [self.connector])
        self.assertEqual(len(self.endpoint.attempts), 1)
        self.assertEqual(len(self.factory.built), 0)
        transport = self.connectionSucceeds()
        self.assertEqual(len(self.factory.built), 1)
        made = transport.protocol.made
        self.assertEqual(len(made), 1)
        self.assertIdentical(made[0], transport)


    def test_connectionLost(self):
        """
        When the connection is lost, both the protocol and the factory will be
        notified via C{connectionLost} and C{clientConnectionLost}.
        """
        why = Failure(RuntimeError())
        proto = self.connectionSucceeds().protocol
        proto.connectionLost(why)
        self.assertEquals(len(self.factory.built), 1)
        self.assertEquals(self.factory.built[0].protocol.lost, [why])
        self.assertEquals(len(self.factory.lost), 1)
        self.assertIdentical(self.factory.lost[0].reason, why)


    def test_connectionFailed(self):
        """
        When the L{Deferred} from the endpoint fails, the L{ClientFactory} gets
        notified via C{clientConnectionFailed}.
        """
        why = Failure(RuntimeError())
        self.connectionFails(why)
        self.assertEquals(len(self.factory.fails), 1)
        self.assertIdentical(self.factory.fails[0].reason, why)


    def test_disconnectWhileConnecting(self):
        """
        When the L{IConnector} is told to C{disconnect} before an in-progress
        L{Deferred} from C{connect} has fired, it will cancel that L{Deferred}.
        """
        self.connector.disconnect()
        self.assertEqual(len(self.factory.fails), 1)
        self.assertTrue(self.factory.fails[0].reason.check(CancelledError))


    def test_disconnectWhileConnected(self):
        """
        When the L{IConnector} is told to C{disconnect} while an existing
        connection is established, that connection will be dropped via
        C{loseConnection}.
        """
        transport = self.connectionSucceeds()
        self.factory.starts[0].disconnect()
        self.assertEqual(transport.lose, [transport])


    def test_connectAfterFailure(self):
        """
        When the L{IConnector} is told to C{connect} after a connection attempt
        has failed, a new connection attempt is started.
        """
        why = Failure(ZeroDivisionError())
        self.connectionFails(why)
        self.connector.connect()
        self.assertEqual(len(self.factory.starts), 2)
        self.assertEqual(len(self.endpoint.attempts), 2)
        self.connectionSucceeds()


    def test_reConnectTooSoon(self):
        """
        When the L{IConnector} is told to C{connect} while another attempt is
        still in flight, it synchronously raises L{RuntimeError}.
        """
        self.assertRaises(RuntimeError, self.connector.connect)
        self.assertEqual(len(self.factory.starts), 1)
        self.assertEqual(len(self.endpoint.attempts), 1)


    def test_stopConnectingWhileConnecting(self):
        """
        When the L{IConnector} is told to C{stopConnecting} while another
        attempt is still in flight, it cancels that connection.
        """
        self.connector.stopConnecting()
        self.assertEqual(len(self.factory.fails), 1)
        self.assertTrue(self.factory.fails[0].reason.check(CancelledError))


    def test_stopConnectingWhileConnected(self):
        """
        When the L{IConnector} is told to C{stopConnecting} while already
        connected, it raises a L{RuntimeError}.
        """
        self.connectionSucceeds()
        self.assertRaises(RuntimeError, self.connector.stopConnecting)


    def test_stopConnectingWhileNotConnected(self):
        """
        When the L{IConnector} is told to C{stopConnecting} while it is not
        connected or connecting, it raises L{RuntimeError}.
        """
        self.connectionFails(Failure(ZeroDivisionError()))
        self.assertRaises(RuntimeError, self.connector.stopConnecting)



##
# Copyright (c) 2013-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from __future__ import print_function

"""
Generic directory service base implementation tests.
"""

from uuid import UUID
from textwrap import dedent

from zope.interface.verify import verifyObject, BrokenMethodImplementation

from twisted.python.constants import (
    Names, NamedConstant, Values, ValueConstant, Flags, FlagConstant
)
from twisted.trial import unittest
from twisted.internet.defer import inlineCallbacks
from twisted.internet.defer import succeed

from ..idirectory import (
    InvalidDirectoryRecordError,
    QueryNotSupportedError, NotAllowedError,
    RecordType, FieldName,
    IDirectoryService, IDirectoryRecord,
)
from ..expression import CompoundExpression, Operand
from ..directory import DirectoryService, DirectoryRecord
from ..util import ConstantsContainer



class TestDirectoryService(DirectoryService):
    recordType = ConstantsContainer((RecordType,))



class StubDirectoryService(TestDirectoryService):
    """
    Stub directory service with some built-in records and an implementation
    of C{recordsFromNonCompoundExpression}.
    """

    def __init__(self, realmName):
        super(StubDirectoryService, self).__init__(realmName)

        self.records = RecordStorage(self, DirectoryRecord)
        self.seenExpressions = []


    def recordsFromExpression(self, expression, recordTypes=None, records=None):
        self.seenExpressions.append(expression)
        return (
            super(StubDirectoryService, self)
            .recordsFromExpression(expression, recordTypes=recordTypes)
        )


    def recordsFromNonCompoundExpression(
        self, expression, recordTypes=None, records=None,
        limitResults=None, timeoutSeconds=None
    ):
        """
        This implementation handles three expressions:

        The expression C{u"None"} will match no records.

        The expressions C{u"twistedmatrix.com"} and C{u"calendarserver.org"}
        will match records that have an email address ending with the
        given expression.
        """
        if expression == u"None":
            return succeed([])

        if expression in (u"twistedmatrix.com", u"calendarserver.org"):
            result = []
            for record in self.records:
                for email in record.emailAddresses:
                    if email.endswith(expression):
                        result.append(record)
                        break
            return succeed(result)

        return (
            super(StubDirectoryService, self)
            .recordsFromNonCompoundExpression(
                expression, recordTypes=recordTypes, records=records
            )
        )



class ServiceMixIn(object):
    """
    MixIn that sets up a service appropriate for testing.
    """

    realmName = u"xyzzy"


    def service(self, subClass=None):
        if subClass is None:
            subClass = self.serviceClass
        return subClass(self.realmName)



class BaseDirectoryServiceTest(ServiceMixIn):
    """
    Tests for L{DirectoryService} and subclasses.
    """

    def test_interface(self):
        """
        Service instance conforms to L{IDirectoryService}.
        """
        service = self.service()
        try:
            verifyObject(IDirectoryService, service)
        except BrokenMethodImplementation as e:
            self.fail(e)


    def test_init(self):
        """
        Test initialization.
        """
        service = self.service()
        self.assertEquals(service.realmName, self.realmName)


    def test_repr(self):
        """
        L{DirectoryService.repr} returns the expected string.
        """
        service = self.service()
        self.assertEquals(
            repr(service),
            "<{0} {1!r}>".format(
                self.serviceClass.__name__, self.realmName
            )
        )


    def test_recordTypes(self):
        """
        L{DirectoryService.recordTypes} returns the supported set of record
        types. For L{DirectoryService}, that's the set of constants in the
        L{DirectoryService.recordType} attribute.
        """
        service = self.service()
        self.assertEquals(
            set(service.recordTypes()),
            set(service.recordType.iterconstants())
        )


    def test_recordsFromNonCompoundExpression_unknownExpression(self):
        """
        L{DirectoryService.recordsFromNonCompoundExpression} with an unknown
        expression type fails with L{QueryNotSupportedError}.
        """
        service = self.service()
        return self.assertFailure(
            service.recordsFromNonCompoundExpression(object()),
            QueryNotSupportedError
        )


    @inlineCallbacks
    def test_recordsFromNonCompoundExpression_emptyRecords(self):
        """
        L{DirectoryService.recordsFromNonCompoundExpression} with an unknown
        expression type and an empty C{records} set returns an empty result.
        """
        service = self.service()
        result = (
            yield service.recordsFromNonCompoundExpression(
                object(), records=()
            )
        )
        self.assertEquals(set(result), set(()))


    def test_recordsFromNonCompoundExpression_nonEmptyRecords(self):
        """
        L{DirectoryService.recordsFromNonCompoundExpression} with an unknown
        expression type and a non-empty C{records} fails with
        L{QueryNotSupportedError}.
        """
        service = self.service()

        wsanchez = self.directoryRecordClass(
            service,
            {
                service.fieldName.recordType: service.recordType.user,
                service.fieldName.uid: u"__wsanchez__",
                service.fieldName.shortNames: [u"wsanchez"],
            }
        )

        return self.assertFailure(
            service.recordsFromNonCompoundExpression(
                object(), records=((wsanchez,))
            ),
            QueryNotSupportedError
        )


    def test_recordsFromExpression_unknownExpression(self):
        """
        L{DirectoryService.recordsFromExpression} with an unknown expression
        type fails with L{QueryNotSupportedError}.
        """
        service = self.service()
        return self.assertFailure(
            service.recordsFromExpression(object()),
            QueryNotSupportedError
        )


    @inlineCallbacks
    def test_recordsFromExpression_emptyExpression(self):
        """
        L{DirectoryService.recordsFromExpression} with an empty
        L{CompoundExpression} returns an empty result.
        """
        service = self.service()

        for operand in Operand.iterconstants():
            result = yield service.recordsFromExpression(
                CompoundExpression((), operand)
            )
            self.assertEquals(set(result), set(()))


    def _unimplemented(self):
        """
        Unimplemented test.
        """
        raise NotImplementedError("Subclasses should implement this test.")


    test_recordWithUID = _unimplemented
    test_recordWithGUID = _unimplemented
    test_recordsWithRecordType = _unimplemented
    test_recordWithShortName = _unimplemented
    test_recordsWithEmailAddress = _unimplemented


    def test_updateRecordsEmpty(self):
        """
        Updating no records is not an error.
        """
        service = self.service()
        for create in (True, False):
            service.updateRecords((), create=create),


    def test_removeRecordsEmpty(self):
        """
        Removing no records is allowed.
        """
        service = self.service()

        service.removeRecords(())



class DirectoryServiceRecordsFromExpressionTest(
    unittest.TestCase, ServiceMixIn
):
    """
    Tests for L{DirectoryService.recordsFromExpression}.
    """

    serviceClass = StubDirectoryService
    directoryRecordClass = DirectoryRecord


    @inlineCallbacks
    def test_recordsFromExpression_single(self):
        """
        L{DirectoryService.recordsFromExpression} handles a single expression.
        """
        service = self.service()

        result = yield service.recordsFromExpression("twistedmatrix.com")

        self.assertEquals(
            set(
                (
                    u"__wsanchez__",
                    u"__glyph__",
                    u"__exarkun__",
                    u"__dreid__",
                )
            ),
            set((record.uid for record in result))
        )


    @inlineCallbacks
    def test_recordsFromExpression_OR(self):
        """
        L{DirectoryService.recordsFromExpression} handles a
        L{CompoundExpression} with L{Operand.OR}.
        """
        service = self.service()

        result = yield service.recordsFromExpression(
            CompoundExpression(
                (
                    u"twistedmatrix.com",
                    u"calendarserver.org",
                ),
                Operand.OR
            )
        )

        self.assertEquals(
            set(
                (
                    u"__wsanchez__",
                    u"__glyph__",
                    u"__sagen__",
                    u"__cdaboo__",
                    u"__dre__",
                    u"__exarkun__",
                    u"__dreid__",
                )
            ),
            set((record.uid for record in result))
        )


    @inlineCallbacks
    def test_recordsFromExpression_AND(self):
        """
        L{DirectoryService.recordsFromExpression} handles a
        L{CompoundExpression} with L{Operand.AND}.
        """
        service = self.service()

        result = yield service.recordsFromExpression(
            CompoundExpression(
                (
                    u"twistedmatrix.com",
                    u"calendarserver.org",
                ),
                Operand.AND
            )
        )

        self.assertEquals(
            set(
                (
                    u"__wsanchez__",
                    u"__glyph__",
                )
            ),
            set((record.uid for record in result))
        )


    @inlineCallbacks
    def test_recordsFromExpression_AND_optimized(self):
        """
        L{DirectoryService.recordsFromExpression} handles a
        L{CompoundExpression} with L{Operand.AND}, and when one of the
        expression matches no records, the subsequent expressions are skipped.
        """
        service = self.service()

        compoundExpression = CompoundExpression(
            (
                u"twistedmatrix.com",
                u"None",
                u"calendarserver.org",
            ),
            Operand.AND
        )

        result = yield service.recordsFromExpression(compoundExpression)

        self.assertEquals(
            set(()),
            set((record.uid for record in result))
        )

        self.assertEquals(
            [compoundExpression, u"twistedmatrix.com", u"None"],
            service.seenExpressions
        )


    def test_recordsFromExpression_unknownOperand(self):
        """
        L{DirectoryService.recordsFromExpression} fails with
        L{QueryNotSupportedError} when given a L{CompoundExpression} with an
        unknown operand.
        """
        service = self.service()

        results = service.recordsFromExpression(
            CompoundExpression(
                (
                    u"twistedmatrix.com",
                    u"calendarserver.org",
                ),
                WackyOperand.WHUH
            )
        )

        return self.assertFailure(results, QueryNotSupportedError)



class DirectoryServiceConvenienceTest(
    unittest.TestCase, BaseDirectoryServiceTest
):
    """
    Tests for L{DirectoryService} convenience methods.
    """

    serviceClass = TestDirectoryService
    directoryRecordClass = DirectoryRecord


    def test_recordWithUID(self):
        """
        L{DirectoryService.recordWithUID} fails with L{QueryNotSupportedError}.
        """
        service = self.service()

        return self.assertFailure(
            service.recordWithUID(u""),
            QueryNotSupportedError
        )


    def test_recordWithGUID(self):
        """
        L{DirectoryService.recordWithGUID} fails with
        L{QueryNotSupportedError}.
        """
        service = self.service()

        return self.assertFailure(
            service.recordWithGUID(UUID(int=0)),
            QueryNotSupportedError
        )


    @inlineCallbacks
    def test_recordsWithRecordType(self):
        """
        L{DirectoryService.recordsWithRecordType} fails with
        L{QueryNotSupportedError}.
        """
        service = self.service()

        for recordType in RecordType.iterconstants():
            yield self.assertFailure(
                service.recordsWithRecordType(recordType),
                QueryNotSupportedError
            )


    @inlineCallbacks
    def test_recordWithShortName(self):
        """
        L{DirectoryService.recordWithShortName} fails with
        L{QueryNotSupportedError}.
        """
        service = self.service()

        for recordType in RecordType.iterconstants():
            yield self.assertFailure(
                service.recordWithShortName(recordType, u""),
                QueryNotSupportedError
            )


    def test_recordsWithEmailAddress(self):
        """
        L{DirectoryService.recordsWithEmailAddress} fails with
        L{QueryNotSupportedError}.
        """
        service = self.service()

        return self.assertFailure(
            service.recordsWithEmailAddress(u"a@b"),
            QueryNotSupportedError
        )



class BaseDirectoryServiceImmutableTest(ServiceMixIn):
    """
    Tests for immutable directory services.
    """

    @inlineCallbacks
    def test_updateRecordsNotAllowed(self):
        """
        Updating records is not allowed.
        """
        service = self.service()

        newRecord = self.directoryRecordClass(
            service,
            fields={
                service.fieldName.uid: u"__plugh__",
                service.fieldName.recordType: service.recordType.user,
                service.fieldName.shortNames: (u"plugh",),
            }
        )

        for create in (True, False):
            yield self.assertFailure(
                service.updateRecords((newRecord,), create=create),
                NotAllowedError,
            )


    def test_removeRecordsNotAllowed(self):
        """
        Removing records is not allowed.
        """
        service = self.service()

        return self.assertFailure(
            service.removeRecords((u"foo",)),
            NotAllowedError,
        )



class DirectoryServiceImmutableTest(
    unittest.TestCase, BaseDirectoryServiceImmutableTest,
):
    """
    Tests for immutable L{DirectoryService}.
    """

    serviceClass = TestDirectoryService
    directoryRecordClass = DirectoryRecord



class BaseDirectoryRecordTest(ServiceMixIn):
    """
    Tests for directory records.
    """

    fields_wsanchez = {
        FieldName.uid: u"UID:wsanchez",
        FieldName.recordType: RecordType.user,
        FieldName.shortNames: (u"wsanchez", u"wilfredo_sanchez"),
        FieldName.fullNames: (
            u"Wilfredo Sanchez",
            u"Wilfredo Sanchez Vega",
        ),
        FieldName.emailAddresses: (
            u"wsanchez@calendarserver.org",
            u"wsanchez@example.com",
        )
    }

    fields_glyph = {
        FieldName.uid: u"UID:glyph",
        FieldName.recordType: RecordType.user,
        FieldName.shortNames: (u"glyph",),
        FieldName.fullNames: (u"Glyph Lefkowitz",),
        FieldName.emailAddresses: (u"glyph@calendarserver.org",)
    }

    fields_sagen = {
        FieldName.uid: u"UID:sagen",
        FieldName.recordType: RecordType.user,
        FieldName.shortNames: (u"sagen",),
        FieldName.fullNames: (u"Morgen Sagen",),
        FieldName.emailAddresses: (u"sagen@CalendarServer.org",)
    }

    fields_nobody = {
        FieldName.uid: u"UID:nobody",
        FieldName.recordType: RecordType.user,
        FieldName.shortNames: (u"nobody",),
    }

    fields_no_shortnames = {
        FieldName.uid: u"UID:no_shortnames",
        FieldName.recordType: RecordType.user,
    }

    fields_empty_shortnames = {
        FieldName.uid: u"UID:empty_shortnames",
        FieldName.recordType: RecordType.user,
        FieldName.shortNames: (),
    }

    fields_none_password = {
        FieldName.uid: u"UID:nobody",
        FieldName.recordType: RecordType.user,
        FieldName.shortNames: (u"nobody",),
        FieldName.password: None,
    }

    fields_staff = {
        FieldName.uid: u"UID:staff",
        FieldName.recordType: RecordType.group,
        FieldName.shortNames: (u"staff",),
        FieldName.fullNames: (u"Staff",),
        FieldName.emailAddresses: (u"staff@CalendarServer.org",)
    }


    def makeRecord(self, fields=None, service=None):
        """
        Create a directory record from fields and a service.

        @param fields: Record fields.
        @type fields: L{dict} with L{FieldName} keys

        @param service: Directory service.
        @type service: L{DirectoryService}

        @return: A directory record.
        @rtype: L{DirectoryRecord}
        """
        if fields is None:
            fields = self.fields_wsanchez
        if service is None:
            service = self.service()
        return self.directoryRecordClass(service, fields)


    def test_interface(self):
        """
        L{DirectoryRecord} complies with L{IDirectoryRecord}.
        """
        record = self.makeRecord()
        try:
            verifyObject(IDirectoryRecord, record)
        except BrokenMethodImplementation as e:
            self.fail(e)


    def test_init(self):
        """
        L{DirectoryRecord} initialization sets service and fields.
        """
        service = self.service()
        wsanchez = self.makeRecord(self.fields_wsanchez, service=service)

        self.assertEquals(wsanchez.service, service)
        self.assertEquals(wsanchez.fields, self.fields_wsanchez)


    def test_initWithNonConstantFieldName(self):
        """
        Directory record field names must be L{NamedConstant}s.
        """
        service = self.service()

        fields = self.fields_wsanchez.copy()
        fields["name"] = u"value"  # Key is not a NamedConstant.

        self.assertRaises(
            InvalidDirectoryRecordError,
            self.makeRecord, fields, service=service
        )


    def test_initWithNoUID(self):
        """
        Directory records must have a UID.
        """
        fields = self.fields_wsanchez.copy()
        del fields[FieldName.uid]
        self.assertRaises(InvalidDirectoryRecordError, self.makeRecord, fields)

        fields = self.fields_wsanchez.copy()
        fields[FieldName.uid] = u""
        self.assertRaises(InvalidDirectoryRecordError, self.makeRecord, fields)


    def test_initWithNoRecordType(self):
        """
        Directory records must have a record type.
        """
        fields = self.fields_wsanchez.copy()
        del fields[FieldName.recordType]
        self.assertRaises(InvalidDirectoryRecordError, self.makeRecord, fields)

        fields = self.fields_wsanchez.copy()
        fields[FieldName.recordType] = None
        self.assertRaises(InvalidDirectoryRecordError, self.makeRecord, fields)


    def test_initWithBogusRecordType(self):
        """
        Directory records must have a known record type.
        """
        fields = self.fields_wsanchez.copy()
        fields[FieldName.recordType] = object()
        self.assertRaises(InvalidDirectoryRecordError, self.makeRecord, fields)


    def test_initNormalizeEmailLowercase(self):
        """
        Email addresses are normalized to lowercase.
        """
        sagen = self.makeRecord(self.fields_sagen)

        self.assertEquals(
            sagen.fields[FieldName.emailAddresses],
            (u"sagen@calendarserver.org",)
        )


    def test_initWithIncorrectFieldTypes(self):
        """
        Raise L{TypeError} if fields are of the wrong type.
        """
        self.assertRaises(
            InvalidDirectoryRecordError,
            self.makeRecord,
            {
                FieldName.uid: "UID:wsanchez",  # Not unicode.
                FieldName.recordType: RecordType.user,
                FieldName.shortNames: (u"wsanchez",),
            }
        )

        self.assertRaises(
            InvalidDirectoryRecordError,
            self.makeRecord,
            {
                FieldName.uid: u"UID:wsanchez",
                FieldName.recordType: RecordType.user,
                FieldName.shortNames: ("wsanchez",),  # Not unicode.
            }
        )


    def test_noneIsAllowed(self):
        """
        Verify a value of None is allowed
        """
        service = self.service()
        record = self.makeRecord(self.fields_none_password, service=service)

        self.assertEquals(record.password, None)


    def _test_containerClassFieldType(self, callback):
        """
        Scaffold for test_initWithContainerClassFieldType_*.
        """

        class ConstantHavingDirectoryService(self.serviceClass):
            fieldName = ConstantsContainer((
                self.serviceClass.fieldName, ConstantHavingFieldName
            ))

        service = self.service(subClass=ConstantHavingDirectoryService)
        fieldName = service.fieldName

        baseFields = {
            FieldName.uid: u"UID:sam",
            FieldName.recordType: RecordType.user,
            FieldName.shortNames: (u"sam",),
        }

        for fieldName, validValue, almostValidValue in (
            (fieldName.eyeColor, Color.blue, OtherColor.mauve),
            (fieldName.language, Language.English, OtherLanguage.French),
            (fieldName.access, Access.read, OtherAccess.delete),
        ):
            fields = baseFields.copy()
            callback(service, fields, fieldName, validValue, almostValidValue)


    def test_initWithContainerClassFieldType_valid(self):
        """
        If C{valueType} is L{Names}, L{Values} or L{Flags}, the expected type
        is L{NamedConstant}, L{ValueConstant} or L{FlagConstant}, respectively.
        Check that these can be used as fields.
        """
        def callback(service, fields, fieldName, validValue, almostValidValue):
            fields.update({fieldName: validValue})
            record = self.makeRecord(fields=fields, service=service)
            self.assertEquals(record.fields[fieldName], validValue)

        self._test_containerClassFieldType(callback)


    def test_initWithContainerClassFieldType_invalid(self):
        """
        If C{valueType} is L{Names}, L{Values} or L{Flags}, the expected type
        is L{NamedConstant}, L{ValueConstant} or L{FlagConstant}, respectively.
        Check that other types raise.
        """
        def callback(service, fields, fieldName, validValue, almostValidValue):
            for invalidValue in (u"string", object()):
                fields.update({fieldName: invalidValue})
                self.assertRaises(
                    InvalidDirectoryRecordError,
                    self.makeRecord, fields=fields, service=service
                )

        self._test_containerClassFieldType(callback)


    def test_initWithContainerClassFieldType_almostValid(self):
        """
        If C{valueType} is L{Names}, L{Values} or L{Flags}, the expected type
        is L{NamedConstant}, L{ValueConstant} or L{FlagConstant}, respectively.
        Check that other container types raise.
        """
        def callback(service, fields, fieldName, validValue, almostValidValue):
            fields.update({fieldName: almostValidValue})
            self.assertRaises(
                InvalidDirectoryRecordError,
                self.makeRecord, fields=fields, service=service
            )

        self._test_containerClassFieldType(callback)

    test_initWithContainerClassFieldType_almostValid.todo = (
        "It would be nice if this raised... presently does not"
    )


    def test_reprWithShortName(self):
        """
        L{DirectoryRecord.repr} returns a string using short name.
        """
        wsanchez = self.makeRecord(self.fields_wsanchez)

        self.assertEquals(
            "<DirectoryRecord (user)wsanchez>",
            repr(wsanchez)
        )


    def test_reprWithoutShortName(self):
        """
        L{DirectoryRecord.repr} returns a string using UID.
        """
        wsanchez = self.makeRecord(self.fields_no_shortnames)

        self.assertEquals(
            "<DirectoryRecord UID:no_shortnames>",
            repr(wsanchez)
        )


    def test_reprEmptyShortName(self):
        """
        L{DirectoryRecord.repr} returns a string using UID.
        """
        wsanchez = self.makeRecord(self.fields_empty_shortnames)

        self.assertEquals(
            "<DirectoryRecord UID:empty_shortnames>",
            repr(wsanchez)
        )


    def test_compare(self):
        """
        Comparison of records.
        """
        fields_glyphmod = self.fields_glyph.copy()
        del fields_glyphmod[FieldName.emailAddresses]

        plugh = self.serviceClass(u"plugh")

        wsanchez = self.makeRecord(self.fields_wsanchez)
        wsanchezmod = self.makeRecord(self.fields_wsanchez, service=plugh)
        glyph = self.makeRecord(self.fields_glyph)
        glyphmod = self.makeRecord(fields_glyphmod)

        self.assertEquals(wsanchez, wsanchez)
        self.assertNotEqual(wsanchez, glyph)
        self.assertNotEqual(glyph, glyphmod)  # UID matches, other fields don't
        self.assertNotEqual(glyphmod, wsanchez)
        self.assertNotEqual(wsanchez, wsanchezmod)  # Different service


    def test_compareOtherType(self):
        """
        Comparison of records with other object types.
        """
        wsanchez = self.makeRecord(self.fields_wsanchez)

        self.assertIdentical(wsanchez.__eq__(object()), NotImplemented)
        self.assertIdentical(wsanchez.__ne__(object()), NotImplemented)


    def test_attributeAccess(self):
        """
        Fields can be accessed as attributes.
        """
        wsanchez = self.makeRecord(self.fields_wsanchez)

        self.assertEquals(
            wsanchez.recordType,
            wsanchez.fields[FieldName.recordType]
        )
        self.assertEquals(
            wsanchez.uid,
            wsanchez.fields[FieldName.uid]
        )
        self.assertEquals(
            wsanchez.shortNames,
            wsanchez.fields[FieldName.shortNames]
        )
        self.assertEquals(
            wsanchez.emailAddresses,
            wsanchez.fields[FieldName.emailAddresses]
        )

        self.assertRaises(AttributeError, lambda: wsanchez.fooBarBaz)

        nobody = self.makeRecord(self.fields_nobody)

        self.assertRaises(AttributeError, lambda: nobody.emailAddresses)


    def test_description(self):
        """
        L{DirectoryRecord.description} returns the expected string.
        """
        sagen = self.makeRecord(self.fields_wsanchez)

        self.assertEquals(
            dedent(
                u"""
                DirectoryRecord:
                  UID = UID:wsanchez
                  email addresses = wsanchez@calendarserver.org, wsanchez@example.com
                  full names = Wilfredo Sanchez, Wilfredo Sanchez Vega
                  record type = user
                  short names = wsanchez, wilfredo_sanchez
                """[1:]
            ),
            sagen.description()
        )


    def test_members_group(self):
        """
        Group members for group records.
        """
        raise NotImplementedError("Subclasses should implement this test.")


    @inlineCallbacks
    def test_members_nonGroup(self):
        """
        Group members for non-group records.  Non-groups have no members.
        """
        wsanchez = self.makeRecord(self.fields_wsanchez)

        self.assertEquals(
            set((yield wsanchez.members())),
            set()
        )


    def test_memberships(self):
        """
        Group memberships.
        """
        raise NotImplementedError("Subclasses should implement this test.")


    @inlineCallbacks
    def test_verifyPlaintextPassword(self):
        """
        Plain text authentication.
        """
        password = u"secret"

        fields = self.fields_wsanchez.copy()
        fields[FieldName.password] = password

        wsanchez = self.makeRecord(fields)

        self.assertTrue((yield wsanchez.verifyPlaintextPassword(password)))
        self.assertFalse((yield wsanchez.verifyPlaintextPassword(u"bleargh")))


    @inlineCallbacks
    def test_verifyHTTPDigest(self):
        """
        HTTP digest authentication.
        """
        password = u"secret"

        fields = self.fields_wsanchez.copy()
        fields[FieldName.password] = password

        wsanchez = self.makeRecord(fields)

        result = yield wsanchez.verifyHTTPDigest(
            username=u"wsanchez",
            realm=None,
            nonce=None,
            cnonce=None,
            algorithm=None,
            nc=None,
            qop=None,
            response=None,
        )

        self.assertTrue(result)
        raise NotImplementedError()

    test_verifyHTTPDigest.todo = "unimplemented"



class DirectoryRecordTest(unittest.TestCase, BaseDirectoryRecordTest):
    """
    Tests for L{DirectoryRecord}.
    """

    serviceClass = TestDirectoryService
    directoryRecordClass = DirectoryRecord


    def test_members_group(self):
        staff = self.makeRecord(self.fields_staff)

        return self.assertFailure(staff.members(), NotImplementedError)


    def test_memberships(self):
        wsanchez = self.makeRecord(self.fields_wsanchez)

        return self.assertFailure(wsanchez.groups(), NotImplementedError)



class RecordStorage(object):
    """
    Container for directory records.
    """

    def __init__(self, service, recordClass):
        self.service = service
        self.recordClass = recordClass
        self.records = []

        self.addDefaultRecords()


    def addDefaultRecords(self):
        """
        Add a known set of records to this service.
        """
        self.addUser(
            shortNames=[u"wsanchez", u"wilfredo_sanchez"],
            fullNames=[
                u"Wilfredo S\xe1nchez Vega",
                u"Wilfredo Sanchez Vega",
                u"Wilfredo Sanchez",
            ],
            emailAddresses=[
                u"wsanchez@bitbucket.calendarserver.org",
                u"wsanchez@devnull.twistedmatrix.com",
            ],
        )

        self.addUser(
            shortNames=[u"glyph"],
            fullNames=[u"Glyph Lefkowitz"],
            emailAddresses=[
                u"glyph@bitbucket.calendarserver.org",
                u"glyph@devnull.twistedmatrix.com",
            ],
        )

        self.addUser(
            shortNames=[u"sagen"],
            fullNames=[u"Morgen Sagen"],
            emailAddresses=[
                u"sagen@bitbucket.calendarserver.org",
                u"shared@example.com",
            ],
        )

        self.addUser(
            shortNames=[u"cdaboo"],
            fullNames=[u"Cyrus Daboo"],
            emailAddresses=[
                u"cdaboo@bitbucket.calendarserver.org",
            ],
        )

        self.addUser(
            shortNames=[u"dre"],
            fullNames=[u"Andre LaBranche"],
            emailAddresses=[
                u"dre@bitbucket.calendarserver.org",
                u"shared@example.com",
            ],
        )

        self.addUser(
            shortNames=[u"exarkun"],
            fullNames=[u"Jean-Paul Calderone"],
            emailAddresses=[
                u"exarkun@devnull.twistedmatrix.com",
            ],
        )

        self.addUser(
            shortNames=[u"dreid"],
            fullNames=[u"David Reid"],
            emailAddresses=[
                u"dreid@devnull.twistedmatrix.com",
            ],
        )

        self.addUser(
            shortNames=[u"joe"],
            fullNames=[u"Joe Schmoe"],
            emailAddresses=[
                u"joe@example.com",
            ],
        )

        self.addUser(
            shortNames=[u"alyssa"],
            fullNames=[u"Alyssa P. Hacker"],
            emailAddresses=[
                u"alyssa@example.com",
            ],
        )

        self.addGroup(
            shortNames=[u"calendar-dev"],
            fullNames=[u"Calendar Developers"],
            emailAddresses=[
                u"calendar-dev@example.com",
            ],
        )

        self.addGroup(
            shortNames=[u"developers"],
            fullNames=[u"Developers"],
            emailAddresses=[
                u"developers@example.com",
            ],
        )


    def addUser(self, shortNames, fullNames, emailAddresses=[]):
        """
        Add a user record with the given field information.

        @param shortNames: Record short names.
        @type shortNames: L{list} of L{unicode}s

        @param fullNames: Record full names.
        @type fullNames: L{list} of L{unicode}s

        @param emailAddresses: Record email addresses.
        @type emailAddresses: L{list} of L{unicode}s
        """
        service = self.service
        fieldName = service.fieldName
        recordType = service.recordType

        self.records.append(self.recordClass(self.service, {
            fieldName.recordType: recordType.user,
            fieldName.uid: u"__{0}__".format(shortNames[0]),
            fieldName.shortNames: shortNames,
            fieldName.fullNames: fullNames,
            fieldName.password: u"".join(reversed(shortNames[0])),
            fieldName.emailAddresses: emailAddresses,
        }))


    def addGroup(self, shortNames, fullNames, emailAddresses=[]):
        """
        Add a group record with the given field information.

        @param shortNames: Record short names.
        @type shortNames: L{list} of L{unicode}s

        @param fullNames: Record full names.
        @type fullNames: L{list} of L{unicode}s

        @param emailAddresses: Record email addresses.
        @type emailAddresses: L{list} of L{unicode}s
        """
        service = self.service
        fieldName = service.fieldName
        recordType = service.recordType

        self.records.append(self.recordClass(self.service, {
            fieldName.recordType: recordType.group,
            fieldName.uid: u"__{0}__".format(shortNames[0]),
            fieldName.shortNames: shortNames,
            fieldName.fullNames: fullNames,
            fieldName.emailAddresses: emailAddresses,
        }))


    def __iter__(self):
        return iter(self.records)



class WackyOperand(Names):
    """
    Wacky operands.
    """
    WHUH = NamedConstant()



class Color(Names):
    """
    Some colors.
    """
    red = NamedConstant()
    green = NamedConstant()
    blue = NamedConstant()
    black = NamedConstant()



class OtherColor(Names):
    """
    More colors.
    """
    mauve = NamedConstant()



class Language(Values):
    """
    Some languages.
    """
    English = ValueConstant(u"en")
    Spanish = ValueConstant(u"sp")



class OtherLanguage(Values):
    """
    More languages.
    """
    French = ValueConstant(u"fr")



class Access(Flags):
    """
    Some access types.
    """
    read = FlagConstant()
    write = FlagConstant()



class OtherAccess(Flags):
    """
    More access types.
    """
    delete = FlagConstant()



class ConstantHavingFieldName(Names):
    """
    Field names with constants values.
    """
    eyeColor = NamedConstant()
    eyeColor.valueType = Color

    language = NamedConstant()
    language.valueType = Language

    access = NamedConstant()
    access.valueType = Access
